# metrics.py
import logging

import keras
import numpy as np
import tensorflow as tf
import tensorflow.keras.backend as K

from sklearn.metrics import roc_curve, auc, average_precision_score

from tensorflow.keras.losses import binary_crossentropy, categorical_crossentropy, sparse_categorical_crossentropy
from tensorflow.keras.losses import LogCosh, CosineSimilarity, MSE, MAE, MAPE, Dice
from keras.saving import register_keras_serializable


STRING_METRICS = [
    'categorical_crossentropy','binary_crossentropy','mean_absolute_error','mae',
    'mean_squared_error', 'mse', 'cosine_similarity', 'log_cosh', 'sparse_categorical_crossentropy',
]


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~ Metrics ~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def weighted_crossentropy(weights, name='anonymous'):
    """A weighted version of tensorflow.keras.objectives.categorical_crossentropy

    Arguments:
        weights = np.array([0.5,2,10]) # Class one at 0.5, class 2 twice the normal weights, class 3 10x.
        name: string identifying the loss to differentiate when models have multiple losses

    Returns:
        keras loss function named name+'_weighted_loss'

    """
    string_globe = 'global ' + name + '_weights\n'
    string_globe += 'global ' + name + '_kweights\n'
    string_globe += name + '_weights = np.array(weights)\n'
    string_globe += name + '_kweights = K.variable('+name+'_weights)\n'
    exec(string_globe, globals(), locals())
    fxn_postfix = '_weighted_loss'
    string_fxn = 'def ' + name + fxn_postfix + '(y_true, y_pred):\n'
    string_fxn += '\ty_pred /= K.sum(y_pred, axis=-1, keepdims=True)\n'
    string_fxn += '\ty_pred = K.clip(y_pred, K.epsilon(), 1 - K.epsilon())\n'
    string_fxn += '\tloss = y_true * K.log(y_pred) * ' + name + '_kweights\n'
    string_fxn += '\tloss = -K.sum(loss, -1)\n'
    string_fxn += '\treturn loss\n'
    exec(string_fxn, globals(), locals())
    loss_fxn = eval(name + fxn_postfix, globals(), locals())
    loss_fxn = register_keras_serializable()(loss_fxn)
    return loss_fxn


def sparse_cross_entropy(window_size: int):
    def _sparse_cross_entropy(y_true, y_pred):
        y_true = tf.reshape(y_true, shape=(-1, window_size))
        loss = tf.keras.losses.SparseCategoricalCrossentropy(
            from_logits=True, reduction='none',
        )(y_true, y_pred)
        return tf.reduce_mean(loss)
    return _sparse_cross_entropy


def angle_between_batches(tensors):
    l0 = K.sqrt(K.sum(K.square(tensors[0]), axis=-1, keepdims=True) + K.epsilon())
    l1 = K.sqrt(K.sum(K.square(tensors[1]), axis=-1, keepdims=True) + K.epsilon())
    numerator = K.sum(tensors[0]*tensors[1], axis=-1, keepdims=True)
    return tf.acos(numerator / (l0*l1))


def two_batch_euclidean(tensors):
    return K.sqrt(K.sum(K.square(tensors[0] - tensors[1]), axis=-1, keepdims=True) + K.epsilon())


def custom_loss_keras(user_id, encodings):
    pairwise_diff = K.expand_dims(encodings, 0) - K.expand_dims(encodings, 1)
    pairwise_squared_distance = K.sum(K.square(pairwise_diff), axis=-1)
    pairwise_distance = K.sqrt(pairwise_squared_distance + K.epsilon())

    user_id = K.squeeze(user_id, axis=1)  # remove the axis added by Keras
    pairwise_equal = K.equal(K.expand_dims(user_id, 0), K.expand_dims(user_id, 1))

    pos_neg = K.cast(pairwise_equal, K.floatx()) * 2 - 1
    return K.sum(pairwise_distance * pos_neg, axis=-1) / 2


def euclid_dist(v):
    return (v[0] - v[1])**2


def angle_between_batches(tensors):
    l0 = K.sqrt(K.sum(K.square(tensors[0]), axis=-1, keepdims=True) + K.epsilon())
    l1 = K.sqrt(K.sum(K.square(tensors[1]), axis=-1, keepdims=True) + K.epsilon())
    numerator = K.sum(tensors[0]*tensors[1], axis=-1, keepdims=True)
    return tf.acos(numerator / (l0*l1))


def paired_angle_between_batches(tensors):
    l0 = K.sqrt(K.sum(K.square(tensors[0]), axis=-1, keepdims=True) + K.epsilon())
    l1 = K.sqrt(K.sum(K.square(tensors[1]), axis=-1, keepdims=True) + K.epsilon())
    numerator = K.sum(tensors[0]*tensors[1], axis=-1, keepdims=True)
    angle_w_self = tf.acos(numerator / (l0*l1))
    # This is very hacky! we assume batch sizes are odd and reverse the batch to compare to others.
    l1_other = K.sqrt(K.sum(K.square(K.reverse(tensors[1], 0)), axis=-1, keepdims=True) + K.epsilon())
    other_numerator = K.sum(tensors[0]*K.reverse(tensors[1], 0), axis=-1, keepdims=True)
    angle_w_other = tf.acos(other_numerator / (l0*l1_other))
    return angle_w_self - angle_w_other


def ignore_zeros_l2(y_true, y_pred):
    mask = K.cast(K.not_equal(y_true, 0), K.floatx())
    return MSE(y_true * mask, y_pred * mask)


def ignore_zeros_logcosh(y_true, y_pred):
    mask = K.cast(K.not_equal(y_true, 0), K.floatx())
    return LogCosh(y_true * mask, y_pred * mask)


def sentinel_logcosh_loss(sentinel: float):
    def ignore_sentinel_logcosh(y_true, y_pred):
        mask = K.cast(K.not_equal(y_true, sentinel), K.floatx())
        return LogCosh(y_true * mask, y_pred * mask)
    return ignore_sentinel_logcosh


def y_true_times_mse(y_true, y_pred):
    return K.maximum(y_true, 1.0)*MSE(y_true, y_pred)


def mse_10x(y_true, y_pred):
    return 10.0*MSE(y_true, y_pred)


def y_true_squared_times_mse(y_true, y_pred):
    return K.maximum(1.0+y_true, 1.0)*K.maximum(1.0+y_true, 1.0)*MSE(y_true, y_pred)


def y_true_cubed_times_mse(y_true, y_pred):
    return K.maximum(y_true, 1.0)*K.maximum(y_true, 1.0)*K.maximum(y_true, 1.0)*MSE(y_true, y_pred)


def y_true_squared_times_logcosh(y_true, y_pred):
    return K.maximum(1.0+y_true, 1.0)*K.maximum(1.0+y_true, 1.0)*LogCosh(y_true, y_pred)


def two_batch_euclidean(tensors):
    return K.sqrt(K.sum(K.square(tensors[0] - tensors[1]), axis=-1, keepdims=True) + K.epsilon())


def custom_loss_keras(user_id, encodings):
    pairwise_diff = K.expand_dims(encodings, 0) - K.expand_dims(encodings, 1)
    pairwise_squared_distance = K.sum(K.square(pairwise_diff), axis=-1)
    pairwise_distance = K.sqrt(pairwise_squared_distance + K.epsilon())

    user_id = K.squeeze(user_id, axis=1)  # remove the axis added by Keras
    pairwise_equal = K.equal(K.expand_dims(user_id, 0), K.expand_dims(user_id, 1))

    pos_neg = K.cast(pairwise_equal, K.floatx()) * 2 - 1
    return K.sum(pairwise_distance * pos_neg, axis=-1) / 2


def pearson(y_true, y_pred):
    # normalizing stage - setting a 0 mean.
    y_true -= K.mean(y_true, axis=-1)
    y_pred -= K.mean(y_pred, axis=-1)
    # normalizing stage - setting a 1 variance
    y_true = K.l2_normalize(y_true, axis=-1)
    y_pred = K.l2_normalize(y_pred, axis=-1)
    # final result
    pearson_correlation = K.sum(y_true * y_pred, axis=-1)
    return pearson_correlation


def abs_pearson(y_true, y_pred):
    # normalizing stage - setting a 0 mean.
    y_true -= K.mean(y_true, axis=-1)
    y_pred -= K.mean(y_pred, axis=-1)
    # normalizing stage - setting a 1 variance
    y_true = K.l2_normalize(y_true, axis=-1)
    y_pred = K.l2_normalize(y_pred, axis=-1)
    # final result
    pearson_correlation = K.sum(y_true * y_pred, axis=-1)
    return tf.math.abs(pearson_correlation)


def _make_riskset(follow_up_times):
    # sort in descending order
    follow_up_times_np = tf.make_ndarray(tf.make_tensor_proto(follow_up_times))
    o = np.argsort(-follow_up_times_np)
    n_samples = follow_up_times_np.shape[0]
    risk_set = np.zeros((n_samples, n_samples))

    for i_start, i_sort in enumerate(o):
        time_i_start = follow_up_times_np[i_sort]
        k = i_start
        while k < n_samples and time_i_start <= follow_up_times_np[o[k]]:
            k += 1
        risk_set[i_sort, o[:k]] = True
    risk_set_tf = tf.convert_to_tensor(risk_set)
    return risk_set_tf


def _softmax_masked(risk_scores, mask, axis=0, keepdims=None):
    """Compute logsumexp across `axis` for entries where `mask` is true."""
    mask_f = K.cast(mask, risk_scores.dtype)
    risk_scores_masked = risk_scores * mask_f
    # for numerical stability, subtract the maximum value before taking the exponential
    amax = K.max(risk_scores_masked, axis=axis, keepdims=True)
    risk_scores_shift = risk_scores_masked - amax

    exp_masked = K.exp(risk_scores_shift) * mask_f
    exp_sum = K.sum(exp_masked, axis=axis, keepdims=True)
    output = amax + K.log(exp_sum)
    if not keepdims:
        output = K.squeeze(output, axis=axis)
    return output


@tf.function
def cox_hazard_loss(y_true, y_pred):
    # move batch dimension to the end so predictions get broadcast row-wise when multiplying by riskset
    pred_t = K.transpose(y_pred[:, 0])
    events = y_true[:, 0]
    follow_up_times = y_true[:, 1]
    # compute log of sum over risk set for each row
    rr = _softmax_masked(pred_t, _make_riskset(follow_up_times), axis=1, keepdims=True)

    losses = events * (rr - y_pred[:, 0])
    loss = K.mean(losses)
    return loss


def survival_likelihood_loss(n_intervals):
    """Create custom Keras loss function for neural network survival model.

    This function is tightly coupled with the function _survival_tensor defined in tensor_from_file.py which builds the y_true tensor.

    Arguments
        n_intervals: the number of survival time intervals
    Returns
        Custom loss function that can be used with Keras
    """

    def loss(y_true, y_pred):
        """
        To play nicely with the Keras framework y_pred is the same shape as y_true.
        However we only consider the first half (n_intervals) of y_pred.
        Arguments
            y_true: Tensor.
              First half of the values are 1 if individual survived that interval, 0 if not.
              Second half of the values are for individuals who failed, and are 1 for time interval during which failure occurred, 0 for other intervals.
              For example given n_intervals = 3 a sample with prevalent disease will have y_true [0, 0, 0, 1, 0, 0]
              a sample with incident disease occurring in the last time bin will have y_true [1, 1, 0, 0, 0, 1]
              a sample who is lost to follow up (censored) in middle time bin will have y_true [1, 0, 0, 0, 0, 0]
            y_pred: Tensor, predicted survival probability (1-hazard probability) for each time interval.
        Returns
            Vector of losses for this minibatch.
        """
        failure_likelihood = 1. - (y_true[:, n_intervals:] * y_pred[:, 0:n_intervals])  # Loss only for individuals who failed
        survival_likelihood = y_true[:, 0:n_intervals] * y_pred[:, 0:n_intervals]  # Loss for intervals that were survived
        survival_likelihood += 1. - y_true[:, 0:n_intervals]  # No survival loss if interval was censored or failed
        return K.sum(-K.log(K.clip(K.concatenate((survival_likelihood, failure_likelihood)), K.epsilon(), None)), axis=-1)  # return -log likelihood

    return loss

def dice(y_true, y_pred):
    return Dice()(y_true, y_pred)
    return Dice(laplace_smoothing=1e-05).mean_loss(y_true, y_pred)

def per_class_dice(labels):
    dice_fxns = []
    for label_key in labels:
        label_idx = labels[label_key]
        fxn_name = label_key.replace('-', '_').replace(' ', '_')
        string_fxn = 'def ' + fxn_name + '_dice(y_true, y_pred):\n'
        string_fxn += '\tdice = tf.keras.losses.Dice()(y_true, y_pred)\n'
        #string_fxn += '\tdice = K.mean(dice, axis=0)['+str(label_idx)+']\n'
        string_fxn += '\treturn dice'

        exec(string_fxn)
        dice_fxn = eval(fxn_name + '_dice')
        dice_fxn = register_keras_serializable()(dice_fxn)
        dice_fxns.append(dice_fxn)

    return dice_fxns

def euclid_dist(v):
    return (v[0] - v[1])**2


def per_class_recall(labels):
    recall_fxns = []
    for label_key in labels:
        label_idx = labels[label_key]
        fxn_name = label_key.replace('-', '_').replace(' ', '_')
        string_fxn = 'def ' + fxn_name + '_recall(y_true, y_pred):\n'
        string_fxn += '\ttrue_positives = K.sum(K.round(K.clip(y_true*y_pred, 0, 1)), axis=0)\n'
        string_fxn += '\tpossible_positives = K.sum(K.round(K.clip(y_true, 0, 1)), axis=0)\n'
        string_fxn += '\treturn true_positives['+str(label_idx)+'] / (possible_positives['+str(label_idx)+'] + K.epsilon())\n'

        exec(string_fxn)
        recall_fxn = eval(fxn_name + '_recall')
        recall_fxn = register_keras_serializable()(recall_fxn)
        recall_fxns.append(recall_fxn)

    return recall_fxns


def per_class_precision(labels):
    precision_fxns = []

    for label_key in labels:
        label_idx = labels[label_key]
        fxn_name = label_key.replace('-', '_').replace(' ', '_')
        string_fxn = 'def ' + fxn_name + '_precision(y_true, y_pred):\n'
        string_fxn += '\ttrue_positives = K.sum(K.round(K.clip(y_true*y_pred, 0, 1)), axis=0)\n'
        string_fxn += '\tpredicted_positives = K.sum(K.round(K.clip(y_pred, 0, 1)), axis=0)\n'
        string_fxn += '\treturn true_positives['+str(label_idx)+'] / (predicted_positives['+str(label_idx)+'] + K.epsilon())\n'

        exec(string_fxn)
        precision_fxn = eval(fxn_name + '_precision')
        precision_fxn = register_keras_serializable()(precision_fxn)
        precision_fxns.append(precision_fxn)

    return precision_fxns


def per_class_recall_3d(labels):
    recall_fxns = []

    for label_key in labels:
        label_idx = labels[label_key]
        fxn_prefix = label_key.replace('-', '_').replace(' ', '_')
        string_fxn = 'def ' + fxn_prefix + '_recall(y_true, y_pred):\n'
        string_fxn += '\ttrue_positives = K.sum(K.sum(K.round(K.clip(y_true*y_pred, 0, 1)), axis=0), axis=0)\n'
        string_fxn += '\tpossible_positives = K.sum(K.sum(K.round(K.clip(y_true, 0, 1)), axis=0), axis=0)\n'
        string_fxn += '\treturn true_positives['+str(label_idx)+'] / (possible_positives['+str(label_idx)+'] + K.epsilon())\n'

        exec(string_fxn)
        recall_fxn = eval(fxn_prefix + '_recall')
        recall_fxn = register_keras_serializable()(recall_fxn)
        recall_fxns.append(recall_fxn)

    return recall_fxns


def per_class_precision_3d(labels):
    precision_fxns = []

    for label_key in labels:
        label_idx = labels[label_key]
        fxn_prefix = label_key.replace('-', '_').replace(' ', '_')
        string_fxn = 'def ' + fxn_prefix + '_precision(y_true, y_pred):\n'
        string_fxn += '\ttrue_positives = K.sum(K.sum(K.round(K.clip(y_true*y_pred, 0, 1)), axis=0), axis=0)\n'
        string_fxn += '\tpredicted_positives = K.sum(K.sum(K.round(K.clip(y_pred, 0, 1)), axis=0), axis=0)\n'
        string_fxn += '\treturn true_positives['+str(label_idx)+'] / (predicted_positives['+str(label_idx)+'] + K.epsilon())\n'

        exec(string_fxn)
        precision_fxn = eval(fxn_prefix + '_precision')
        precision_fxn = register_keras_serializable()(precision_fxn)
        precision_fxns.append(precision_fxn)

    return precision_fxns


def per_class_recall_4d(labels):
    recall_fxns = []

    for label_key in labels:
        label_idx = labels[label_key]
        fxn_prefix = label_key.replace('-', '_').replace(' ', '_')
        string_fxn = 'def ' + fxn_prefix + '_recall(y_true, y_pred):\n'
        string_fxn += '\ttrue_positives = K.sum(K.sum(K.sum(K.round(K.clip(y_true*y_pred, 0, 1)), axis=0), axis=0), axis=0)\n'
        string_fxn += '\tpossible_positives = K.sum(K.sum(K.sum(K.round(K.clip(y_true, 0, 1)), axis=0), axis=0), axis=0)\n'
        string_fxn += '\treturn true_positives['+str(label_idx)+'] / (possible_positives['+str(label_idx)+'] + K.epsilon())\n'

        exec(string_fxn)
        recall_fxn = eval(fxn_prefix + '_recall')
        recall_fxn = register_keras_serializable()(recall_fxn)
        recall_fxns.append(recall_fxn)

    return recall_fxns


def per_class_precision_4d(labels):
    precision_fxns = []

    for label_key in labels:
        label_idx = labels[label_key]
        fxn_prefix = label_key.replace('-', '_').replace(' ', '_')
        string_fxn = 'def ' + fxn_prefix + '_precision(y_true, y_pred):\n'
        string_fxn += '\ttrue_positives = K.sum(K.sum(K.sum(K.round(K.clip(y_true*y_pred, 0, 1)), axis=0), axis=0), axis=0)\n'
        string_fxn += '\tpredicted_positives = K.sum(K.sum(K.sum(K.round(K.clip(y_pred, 0, 1)), axis=0), axis=0), axis=0)\n'
        string_fxn += '\treturn true_positives['+str(label_idx)+'] / (predicted_positives['+str(label_idx)+'] + K.epsilon())\n'

        exec(string_fxn)
        precision_fxn = eval(fxn_prefix + '_precision')
        precision_fxn = register_keras_serializable()(precision_fxn)

        precision_fxns.append(precision_fxn)

    return precision_fxns


def per_class_recall_5d(labels):
    recall_fxns = []

    for label_key in labels:
        label_idx = labels[label_key]
        fxn_prefix = label_key.replace('-', '_').replace(' ', '_')
        string_fxn = 'def ' + fxn_prefix + '_recall(y_true, y_pred):\n'
        string_fxn += '\ttrue_positives = K.sum(K.sum(K.sum(K.sum(K.round(K.clip(y_true*y_pred, 0, 1)), axis=0), axis=0), axis=0), axis=0)\n'
        string_fxn += '\tpossible_positives = K.sum(K.sum(K.sum(K.sum(K.round(K.clip(y_true, 0, 1)), axis=0), axis=0), axis=0), axis=0)\n'
        string_fxn += '\treturn true_positives['+str(label_idx)+'] / (possible_positives['+str(label_idx)+'] + K.epsilon())\n'

        exec(string_fxn)
        recall_fxn = eval(fxn_prefix + '_recall')
        recall_fxn = register_keras_serializable()(recall_fxn)
        recall_fxns.append(recall_fxn)

    return recall_fxns


def per_class_precision_5d(labels):
    precision_fxns = []

    for label_key in labels:
        label_idx = labels[label_key]
        fxn_prefix = label_key.replace('-', '_').replace(' ', '_')
        string_fxn = 'def ' + fxn_prefix + '_precision(y_true, y_pred):\n'
        string_fxn += '\ttrue_positives = K.sum(K.sum(K.sum(K.sum(K.round(K.clip(y_true*y_pred, 0, 1)), axis=0), axis=0), axis=0), axis=0)\n'
        string_fxn += '\tpredicted_positives = K.sum(K.sum(K.sum(K.sum(K.round(K.clip(y_pred, 0, 1)), axis=0), axis=0), axis=0), axis=0)\n'
        string_fxn += '\treturn true_positives['+str(label_idx)+'] / (predicted_positives['+str(label_idx)+'] + K.epsilon())\n'

        exec(string_fxn)
        precision_fxn = eval(fxn_prefix + '_precision')
        precision_fxn = register_keras_serializable()(precision_fxn)
        precision_fxns.append(precision_fxn)

    return precision_fxns


def get_metric_dict(output_tensor_maps):
    metrics = {}
    losses = []
    loss_weights = []
    for tm in output_tensor_maps:
        loss_weights.append(tm.loss_weight)
        for m in tm.metrics:
            if isinstance(m, str):
                metrics[m] = m
            elif hasattr(m, '__name__'):
                metrics[m.__name__] = m

        if tm.loss == 'categorical_crossentropy':
            losses.append(categorical_crossentropy)
        elif tm.loss == 'sparse_categorical_crossentropy':
            losses.append(sparse_categorical_crossentropy)
        elif tm.loss == 'binary_crossentropy':
            losses.append(binary_crossentropy)
        elif tm.loss == 'mean_absolute_error' or tm.loss == 'mae':
            losses.append(MSE)
        elif tm.loss == 'mean_squared_error' or tm.loss == 'mse':
            losses.append(MSE)
        elif tm.loss == 'cosine_similarity':
            losses.append(CosineSimilarity)
        elif tm.loss == 'log_cosh':
            losses.append(LogCosh)
        elif tm.loss == 'mape':
            losses.append(MAPE)
        elif hasattr(tm.loss,  '__name__'):
            metrics[tm.loss.__name__] = tm.loss
            losses.append(tm.loss)
        else:
            losses.append(tm.loss)

    def loss(y_true, y_pred):
        my_loss = 0
        for my_loss_fxn, loss_weight in zip(losses, loss_weights):
            my_loss += loss_weight * my_loss_fxn(y_true, y_pred)
        return my_loss
    metrics['loss'] = loss

    return metrics


def get_roc_aucs(predictions, truth, labels):
    """Compute ROC AUC for each label of each given model"""
    aucs = dict()

    for model in predictions.keys():
        roc_auc = dict()
        for label_name in labels.keys():
            label_encoding = labels[label_name]
            fpr, tpr,  _ = roc_curve(truth[:, label_encoding], predictions[model][:, label_encoding])
            roc_auc[label_name] = auc(fpr, tpr)
        aucs[model] = roc_auc

    return aucs


def get_precision_recall_aucs(predictions, truth, labels):
    """Compute Precision-Recall AUC for each label of each given model"""
    aucs = dict()

    for model in predictions.keys():
        average_precision = dict()
        for label_name in labels.keys():
            label_encoding = labels[label_name]
            average_precision[label_name] = average_precision_score(truth[:, label_encoding], predictions[model][:, label_encoding])
        aucs[model] = average_precision

    return aucs


def log_aucs(**aucs):
    """Log and tabulate AUCs given as nested dictionaries in the format '{model: {label: auc}}'"""
    def dashes(n): return '-' * n

    header = "{:<40} {:<28} {:<15}"
    row = "{:<40} {:<28} {:<15.4f}"
    width = 90
    logging.info(dashes(width))

    for auc_name, auc_value in aucs.items():
        logging.info(header.format('Model', 'Label', auc_name+' AUC'))
        triplets = []
        for model, model_value in auc_value.items():
            for label, auc in model_value.items():
                triplets.append((model, label, auc))
        for model, label, auc in sorted(triplets, key=lambda x: x[1]):
            logging.info(row.format(model, label, auc))
        logging.info(dashes(width))


def coefficient_of_determination(truth, predictions, eps=1e-6):
    true_mean = np.mean(truth)
    total_sum_of_squares = np.sum((truth - true_mean) * (truth - true_mean))
    residual_sum_of_squares = np.sum((predictions - truth) * (predictions - truth))
    r_squared = 1 - (residual_sum_of_squares / (total_sum_of_squares + eps))
    return r_squared


def get_pearson_coefficients(predictions, truth):
    coefs = dict()
    for model in predictions.keys():
        # corrcoef() returns full covariance matrix
        pearson = np.corrcoef(predictions[model].flatten(), truth.flatten())[1, 0]
        coefs[model] = pearson

    return coefs


def log_pearson_coefficients(coefs, label):
    def dashes(n): return '-' * n

    header = "{:<30} {:<25} {:<15} {:<15}"
    row = "{:<30} {:<25} {:<15.10f} {:<15.10f}"
    width = 85
    logging.info(dashes(width))
    logging.info(header.format('Model', 'Label', 'Pearson R', 'Pearson R^2'))
    for model, coef in coefs.items():
        logging.info(row.format(model, label, coef, coef*coef))
    logging.info(dashes(width))


def _unpack_truth_into_events(truth, intervals):
    event_time = np.argmin(np.diff(truth[:, :intervals]), axis=-1)
    event_time[truth[:, intervals-1] == 1] = intervals-1  # If the sample is never censored set event time to max time
    event_indicator = np.sum(truth[:, intervals:], axis=-1).astype(bool)
    return event_indicator, event_time


def _get_comparable(event_indicator, event_time, order):
    n_samples = len(event_time)
    tied_time = 0
    comparable = {}
    i = 0
    while i < n_samples - 1:
        time_i = event_time[order[i]]
        start = i + 1
        end = start
        while end < n_samples and event_time[order[end]] == time_i:
            end += 1

        # check for tied event times
        event_at_same_time = event_indicator[order[i:end]]
        censored_at_same_time = ~event_at_same_time
        for j in range(i, end):
            if event_indicator[order[j]]:
                mask = np.zeros(n_samples, dtype=bool)
                mask[end:] = True
                # an event is comparable to censored samples at same time point
                mask[i:end] = censored_at_same_time
                comparable[j] = mask
                tied_time += censored_at_same_time.sum()
        i = end

    return comparable, tied_time


def concordance_index(prediction, truth, tied_tol=1e-8):
    intervals = truth.shape[-1] // 2
    event_indicator, event_time = _unpack_truth_into_events(truth, intervals)
    estimate = np.cumprod(prediction[:, :intervals], axis=-1)[:, -1]
    order = np.argsort(event_time)
    comparable, tied_time = _get_comparable(event_indicator, event_time, order)

    concordant = 0
    discordant = 0
    tied_risk = 0
    numerator = 0.0
    denominator = 0.0
    for ind, mask in comparable.items():
        est_i = estimate[order[ind]]
        event_i = event_indicator[order[ind]]

        est = estimate[order[mask]]

        assert event_i, 'got censored sample at index %d, but expected uncensored' % order[ind]

        ties = np.absolute(est - est_i) <= tied_tol
        n_ties = ties.sum()

        # an event should have a higher score
        con = est > est_i
        n_con = con[~ties].sum()

        numerator += n_con + 0.5 * n_ties
        denominator += mask.sum()

        tied_risk += n_ties
        concordant += n_con
        discordant += est.size - n_con - n_ties

    cindex = numerator / denominator
    return cindex, concordant, discordant, tied_risk, tied_time


def _estimate_concordance_index(event_indicator, event_time, estimate, weights, tied_tol=1e-8):
    order = np.argsort(event_time)
    comparable, tied_time = _get_comparable(event_indicator, event_time, order)
    if len(comparable) == 0:
        raise ValueError("Data has no comparable pairs, cannot estimate concordance index.")

    concordant = 0
    discordant = 0
    tied_risk = 0
    numerator = 0.0
    denominator = 0.0
    for ind, mask in comparable.items():
        est_i = estimate[order[ind]]
        event_i = event_indicator[order[ind]]
        w_i = weights[order[ind]]

        est = estimate[order[mask]]

        assert event_i, 'got censored sample at index %d, but expected uncensored' % order[ind]

        ties = np.absolute(est - est_i) <= tied_tol
        n_ties = ties.sum()
        # an event should have a higher score
        con = est < est_i
        n_con = con[~ties].sum()

        numerator += w_i * n_con + 0.5 * w_i * n_ties
        denominator += w_i * mask.sum()

        tied_risk += n_ties
        concordant += n_con
        discordant += est.size - n_con - n_ties

    cindex = numerator / denominator
    return cindex, concordant, discordant, tied_risk, tied_time


def concordance_index_censored(event_indicator, event_time, estimate, tied_tol=1e-8):
    """Concordance index for right-censored data
    The concordance index is defined as the proportion of all comparable pairs
    in which the predictions and outcomes are concordant.
    Two samples are comparable if (i) both of them experienced an event (at different times),
    or (ii) the one with a shorter observed survival time experienced an event, in which case
    the event-free subject "outlived" the other. A pair is not comparable if they experienced
    events at the same time.
    Concordance intuitively means that two samples were ordered correctly by the model.
    More specifically, two samples are concordant, if the one with a higher estimated
    risk score has a shorter actual survival time.
    When predicted risks are identical for a pair, 0.5 rather than 1 is added to the count
    of concordant pairs.
    See the :ref:`User Guide </user_guide/evaluating-survival-models.ipynb>`
    and [1]_ for further description.
    Parameters
    ----------
    event_indicator : array-like, shape = (n_samples,)
        Boolean array denotes whether an event occurred
    event_time : array-like, shape = (n_samples,)
        Array containing the time of an event or time of censoring
    estimate : array-like, shape = (n_samples,)
        Estimated risk of experiencing an event
    tied_tol : float, optional, default: 1e-8
        The tolerance value for considering ties.
        If the absolute difference between risk scores is smaller
        or equal than `tied_tol`, risk scores are considered tied.
    Returns
    -------
    cindex : float
        Concordance index
    concordant : int
        Number of concordant pairs
    discordant : int
        Number of discordant pairs
    tied_risk : int
        Number of pairs having tied estimated risks
    tied_time : int
        Number of comparable pairs sharing the same time
    See also
    --------
    concordance_index_ipcw
        Alternative estimator of the concordance index with less bias.
    References
    ----------
    .. [1] Harrell, F.E., Califf, R.M., Pryor, D.B., Lee, K.L., Rosati, R.A,
           "Multivariable prognostic models: issues in developing models,
           evaluating assumptions and adequacy, and measuring and reducing errors",
           Statistics in Medicine, 15(4), 361-87, 1996.
    """
    w = np.ones_like(estimate)
    return _estimate_concordance_index(event_indicator, event_time, estimate, w, tied_tol)


class KernelInceptionDistance(keras.metrics.Metric):
    def __init__(self, name, input_shape, kernel_image_size, **kwargs):
        super().__init__(name=name, **kwargs)

        # KID is estimated per batch and is averaged across batches
        self.kid_tracker = keras.metrics.Mean(name="kid_tracker")

        # a pretrained InceptionV3 is used without its classification layer
        # transform the pixel values to the 0-255 range, then use the same
        # preprocessing as during pretraining
        self.encoder = keras.Sequential(
            [
                keras.Input(shape=input_shape), # TODO: handle multi-channel
                keras.layers.Lambda(lambda x: tf.tile(x, [1, 1, 1, 3])),
                keras.layers.Rescaling(255.0),
                keras.layers.Resizing(height=kernel_image_size, width=kernel_image_size),
                keras.layers.Lambda(keras.applications.inception_v3.preprocess_input),
                keras.applications.InceptionV3(
                    include_top=False,
                    input_shape=(kernel_image_size, kernel_image_size, 3),
                    weights="imagenet",
                ),
                keras.layers.GlobalAveragePooling2D(),
            ],
            name="inception_encoder",
        )

    def polynomial_kernel(self, features_1, features_2):
        feature_dimensions = tf.cast(tf.shape(features_1)[1], dtype=tf.float32)
        return (features_1 @ tf.transpose(features_2) / feature_dimensions + 1.0) ** 3.0

    def update_state(self, real_images, generated_images, sample_weight=None):
        real_features = self.encoder(real_images, training=False)
        generated_features = self.encoder(generated_images, training=False)

        # compute polynomial kernels using the two sets of features
        kernel_real = self.polynomial_kernel(real_features, real_features)
        kernel_generated = self.polynomial_kernel(
            generated_features, generated_features
        )
        kernel_cross = self.polynomial_kernel(real_features, generated_features)

        # estimate the squared maximum mean discrepancy using the average kernel values
        batch_size = tf.shape(real_features)[0]
        batch_size_f = tf.cast(batch_size, dtype=tf.float32)
        mean_kernel_real = tf.reduce_sum(kernel_real * (1.0 - tf.eye(batch_size))) / (
            batch_size_f * (batch_size_f - 1.0)
        )
        mean_kernel_generated = tf.reduce_sum(
            kernel_generated * (1.0 - tf.eye(batch_size))
        ) / (batch_size_f * (batch_size_f - 1.0))
        mean_kernel_cross = tf.reduce_mean(kernel_cross)
        kid = mean_kernel_real + mean_kernel_generated - 2.0 * mean_kernel_cross

        # update the average KID estimate
        self.kid_tracker.update_state(kid)

    def result(self):
        return self.kid_tracker.result()

    def reset_state(self):
        self.kid_tracker.reset_state()


class InceptionScore(keras.metrics.Metric):
    def __init__(self, name, input_shape, kernel_image_size, **kwargs):
        super().__init__(name=name, **kwargs)

        # Inception score is estimated per batch and averaged across batches
        self.is_tracker = keras.metrics.Mean(name="is_tracker")

        # A pretrained InceptionV3 is used without its classification layer
        # We preprocess the images as during the pretraining of InceptionV3
        self.encoder = keras.Sequential(
            [
                keras.Input(shape=input_shape),  # TODO: handle multi-channel
                keras.layers.Lambda(lambda x: tf.tile(x, [1, 1, 1, 3])),  # Ensure 3 channels
                keras.layers.Rescaling(255.0),
                keras.layers.Resizing(height=kernel_image_size, width=kernel_image_size),
                keras.layers.Lambda(keras.applications.inception_v3.preprocess_input),
                keras.applications.InceptionV3(
                    include_top=True,  # Include the classification layer for IS
                    input_shape=(kernel_image_size, kernel_image_size, 3),
                    weights="imagenet",
                ),
            ],
            name="inception_encoder",
        )

    def update_state(self, generated_images, sample_weight=None):
        # Get the predicted class probabilities from the InceptionV3 model
        logits = self.encoder(generated_images, training=False)
        softmax_probs = tf.nn.softmax(logits, axis=-1)

        # Compute p(y|x) for each generated image (softmax probabilities)
        p_y_given_x = softmax_probs

        # Compute the marginal distribution p(y) across all generated images
        p_y = tf.reduce_mean(p_y_given_x, axis=0)

        # Compute KL divergence between p(y|x) and p(y) for each image
        kl_divergence = tf.reduce_sum(p_y_given_x * (tf.math.log(p_y_given_x) - tf.math.log(p_y)), axis=-1)

        # Inception score is the exponentiation of the mean KL divergence
        is_score = tf.exp(tf.reduce_mean(kl_divergence))

        # Update the average IS estimate
        self.is_tracker.update_state(is_score)

    def result(self):
        return self.is_tracker.result()

    def reset_state(self):
        self.is_tracker.reset_state()


class MultiScaleSSIM(keras.metrics.Metric):
    def __init__(self, name="multi_scale_ssim", **kwargs):
        super(MultiScaleSSIM, self).__init__(name=name, **kwargs)
        self.total_ssim = self.add_weight(name="total_ssim", initializer="zeros")
        self.count = self.add_weight(name="count", initializer="zeros")

    def update_state(self, y_true, y_pred, max_val, sample_weight=None):
        # Calculate MS-SSIM for the batch
        ssim = tf.image.ssim_multiscale(y_true, y_pred, max_val=max_val, power_factors=[0.25, 0.25, 0.25, 0.25])
        if sample_weight is not None:
            ssim = tf.multiply(ssim, sample_weight)

        # Update total MS-SSIM and count
        self.total_ssim.assign_add(tf.reduce_sum(ssim))
        self.count.assign_add(tf.cast(tf.size(ssim), tf.float32))

    def result(self):
        # Return the mean MS-SSIM over all batches
        return tf.divide(self.total_ssim, self.count)

    def reset_state(self):
        # Reset the metric state variables
        self.total_ssim.assign(0.0)
        self.count.assign(0.0)

        
def _register_all(module_globals):
    for name, obj in module_globals.items():
        if callable(obj) and not name.startswith("_"):
            module_globals[name] = register_keras_serializable()(obj)

_register_all(globals())


import os
import h5py
import logging
import argparse
import traceback
from collections import Counter

from ml4h.defines import TENSOR_EXT, HD5_GROUP_CHAR

"""
This script copies all the hd5 datasets from all hd5 files within the 'sources'
directories to the same-named files within the 'destination' directory.

If the tensor files are not in a local filesystem, they can be downloaded via gsutil:
gsutil -m cp -r <gcs bucket with tensors> <local directory>

If the destination directory and/or file(s) don't exist, it creates them.
If any of the source files contain the same dataset at the same group path, it errors out.

Example command line:
python .merge_hd5s.py \
    --sources /path/to/src/continuous/tensor/directory/ /path/to/src/categorical/tensor/directory/ \
    --destination /path/to/output/directory/
"""


def merge_hd5s_into_destination(destination, sources, min_sample_id, max_sample_id, intersect, inplace):
    stats = Counter()
    if not os.path.exists(os.path.dirname(destination)):
        os.makedirs(os.path.dirname(destination))

    if inplace:
        sample_set = os.listdir(destination)
    elif intersect:
        sample_sets = [os.listdir(source_folder) for source_folder in sources]
        sample_set = set(sample_sets[0]).intersection(*sample_sets[1:])

    for source_folder in sources:
        for source_file in os.listdir(source_folder):
            if not source_file.endswith(TENSOR_EXT):
                continue
            if not (min_sample_id <= int(os.path.splitext(source_file)[0]) < max_sample_id):
                continue
            if (intersect or inplace) and source_file not in sample_set:
                continue

            with h5py.File(os.path.join(destination, source_file), 'a') as destination_hd5:
                with h5py.File(os.path.join(source_folder, source_file), 'r') as source_hd5:
                    _copy_hd5_datasets(source_hd5, destination_hd5, stats=stats)
        logging.info(f"Done copying source folder {source_folder}")

    for k in stats:
        logging.info(f"{k} has {stats[k]} tensors")


def _copy_hd5_datasets(source_hd5, destination_hd5, group_path=HD5_GROUP_CHAR, stats=None):
    for k in source_hd5[group_path]:
        if isinstance(source_hd5[group_path][k], h5py.Dataset):
            try:
                if source_hd5[group_path][k].chunks is None:
                    destination_hd5.create_dataset(group_path + k, data=source_hd5[group_path][k])
                else:
                    destination_hd5.create_dataset(group_path + k, data=source_hd5[group_path][k], compression='gzip')
                stats[group_path + k] += 1
            except (OSError, KeyError, RuntimeError, ValueError) as e:
                logging.debug(f"Error trying to write:{k} at group path:{group_path} error:{e}\n{traceback.format_exc()}\n")
        else:
            logging.debug(f"copying group {group_path + k}")
            _copy_hd5_datasets(source_hd5, destination_hd5, group_path=group_path + k + HD5_GROUP_CHAR, stats=stats)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--sources', nargs='+', help='List of source directories with hd5 files')
    parser.add_argument('--destination', help='Destination directory to copy hd5 datasets to')
    parser.add_argument('--min_sample_id', default=0, type=int, help='Minimum sample id to write to tensor.')
    parser.add_argument('--max_sample_id', default=7000000, type=int, help='Maximum sample id to write to tensor.')
    parser.add_argument("--logging_level", default='INFO', help="Logging level", choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"])
    parser.add_argument('--inplace', default=False, action='store_true', help='Merge into pre-existing destination tensors')
    parser.add_argument(
        '--intersect', default=False, action='store_true',
        help='Only merge files if the sample id is in every source directory (and if destination if destination is not empty',
    )
    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()
    logging.getLogger().setLevel(args.logging_level)
    merge_hd5s_into_destination(args.destination, args.sources, args.min_sample_id, args.max_sample_id, args.intersect, args.inplace)



# recipes.py

# Imports
import os
import csv
import copy
import h5py
import glob
import logging
import datetime
import numpy as np
import pandas as pd

from functools import reduce
from google.cloud import storage
from timeit import default_timer as timer
from collections import Counter, defaultdict

from ml4h.arguments import parse_args
from ml4h.models.inspect import saliency_map
from ml4h.optimizers import find_learning_rate
from ml4h.defines import TENSOR_EXT, MODEL_EXT
from ml4h.tensormap.tensor_map_maker import write_tensor_maps
from ml4h.tensorize.tensor_writer_mgb import write_tensors_mgb
from ml4h.models.model_factory import make_multimodal_multitask_model
from ml4h.ml4ht_integration.tensor_generator import TensorMapDataLoader2
from ml4h.tensor_generators import BATCH_INPUT_INDEX, BATCH_OUTPUT_INDEX, BATCH_PATHS_INDEX
from ml4h.explorations import test_labels_to_label_map, infer_with_pixels, infer_stats_from_segmented_regions
from ml4h.explorations import mri_dates, ecg_dates, predictions_to_pngs, sample_from_language_model
from ml4h.plots import plot_roc, plot_precision_recall_per_class, plot_scatter
from ml4h.explorations import plot_while_learning, plot_histograms_of_tensors_in_pdf, explore, pca_on_tsv
from ml4h.models.legacy_models import get_model_inputs_outputs, make_shallow_model, make_hidden_layer_model
from ml4h.models.train import train_model_from_generators, train_diffusion_model, train_diffusion_control_model
from ml4h.tensor_generators import TensorGenerator, test_train_valid_tensor_generators, big_batch_from_minibatch_generator
from ml4h.data_descriptions import dataframe_data_description_from_tensor_map, ECGDataDescription, DataFrameDataDescription
from ml4h.metrics import get_roc_aucs, get_precision_recall_aucs, get_pearson_coefficients, log_aucs, log_pearson_coefficients, concordance_index_censored

from ml4h.plots import plot_reconstruction, plot_saliency_maps, plot_partners_ecgs, plot_ecg_rest_mp
from ml4h.plots import plot_dice, plot_reconstruction, plot_hit_to_miss_transforms, plot_saliency_maps, plot_partners_ecgs, plot_ecg_rest_mp

from ml4h.plots import subplot_rocs, subplot_comparison_rocs, subplot_scatters, subplot_comparison_scatters, plot_prediction_calibrations
from ml4h.models.legacy_models import embed_model_predict, make_siamese_model, legacy_multimodal_multitask_model
from ml4h.plots import plot_dice, subplot_roc_per_class, plot_tsne, plot_survival
from ml4h.plots import evaluate_predictions, plot_scatters, plot_rocs, plot_precision_recalls
from ml4h.tensorize.tensor_writer_ukbb import write_tensors, append_fields_from_csv, append_gene_csv, write_tensors_from_dicom_pngs, write_tensors_from_ecg_pngs

from ml4ht.data.util.date_selector import DATE_OPTION_KEY
from ml4ht.data.sample_getter import DataDescriptionSampleGetter
from ml4ht.data.data_loader import SampleGetterIterableDataset, shuffle_get_epoch

from torch.utils.data import DataLoader


def run(args):
    start_time = timer()  # Keep track of elapsed execution time
    try:
        if 'tensorize' == args.mode:
            write_tensors(
                args.id, args.xml_folder, args.zip_folder, args.output_folder, args.tensors, args.dicoms, args.mri_field_ids, args.xml_field_ids,
                args.write_pngs, args.min_sample_id, args.max_sample_id, args.min_values,
            )
        elif 'tensorize_pngs' == args.mode:
            write_tensors_from_dicom_pngs(args.tensors, args.dicoms, args.app_csv, args.dicom_series, args.min_sample_id, args.max_sample_id, args.x, args.y)
        elif 'tensorize_ecg_pngs' == args.mode:
            write_tensors_from_ecg_pngs(args.tensors, args.xml_folder, args.min_sample_id, args.max_sample_id)
        elif 'tensorize_partners' == args.mode:
            write_tensors_mgb(args.xml_folder, args.tensors, args.num_workers)
        elif 'explore' == args.mode:
            explore(args)
        elif 'train' == args.mode:
            train_multimodal_multitask(args)
        elif 'train_legacy' == args.mode:
            train_legacy(args)
        elif 'train_xdl' == args.mode:
            train_xdl(args)
        elif 'train_xdl_af' == args.mode:
            train_xdl_af(args)
        elif 'test' == args.mode:
            test_multimodal_multitask(args)
        elif 'compare' == args.mode:
            compare_multimodal_multitask_models(args)
        elif 'infer' == args.mode:
            infer_multimodal_multitask(args)
        elif 'infer_hidden' == args.mode:
            infer_hidden_layer_multimodal_multitask(args)
        elif 'infer_pixels' == args.mode:
            infer_with_pixels(args)
        elif 'infer_stats_from_segmented_regions' == args.mode:
            infer_stats_from_segmented_regions(args)
        elif 'infer_encoders' == args.mode:
            infer_encoders_block_multimodal_multitask(args)
        elif 'infer_xdl' == args.mode:
            infer_xdl(args)
        elif 'test_scalar' == args.mode:
            test_multimodal_scalar_tasks(args)
        elif 'compare_scalar' == args.mode:
            compare_multimodal_scalar_task_models(args)
        elif 'plot_predictions' == args.mode:
            plot_predictions(args)
        elif 'plot_while_training' == args.mode:
            plot_while_training(args)
        elif 'plot_saliency' == args.mode:
            saliency_maps(args)
        elif 'plot_mri_dates' == args.mode:
            mri_dates(args.tensors, args.output_folder, args.id)
        elif 'plot_ecg_dates' == args.mode:
            ecg_dates(args.tensors, args.output_folder, args.id)
        elif 'plot_histograms' == args.mode:
            plot_histograms_of_tensors_in_pdf(args.id, args.tensors, args.output_folder, args.max_samples)
        elif 'plot_resting_ecgs' == args.mode:
            plot_ecg_rest_mp(args.tensors, args.min_sample_id, args.max_sample_id, args.output_folder, args.num_workers)
        elif 'plot_partners_ecgs' == args.mode:
            plot_partners_ecgs(args)
        elif 'train_shallow' == args.mode:
            train_shallow_model(args)
        elif 'train_diffusion' == args.mode:
            train_diffusion_model(args)
        elif 'train_diffusion_control' == args.mode:
            train_diffusion_control_model(args)
        elif 'train_diffusion_supervise' == args.mode:
            train_diffusion_control_model(args, supervised=True)
        elif 'train_siamese' == args.mode:
            train_siamese_model(args)
        elif 'write_tensor_maps' == args.mode:
            write_tensor_maps(args)
        elif 'append_continuous_csv' == args.mode:
            append_fields_from_csv(args.tensors, args.app_csv, 'continuous', ',')
        elif 'append_categorical_csv' == args.mode:
            append_fields_from_csv(args.tensors, args.app_csv, 'categorical', ',')
        elif 'append_continuous_tsv' == args.mode:
            append_fields_from_csv(args.tensors, args.app_csv, 'continuous', '\t')
        elif 'append_categorical_tsv' == args.mode:
            append_fields_from_csv(args.tensors, args.app_csv, 'categorical', '\t')
        elif 'append_gene_csv' == args.mode:
            append_gene_csv(args.tensors, args.app_csv, ',')
        elif 'pca_on_hidden_inference' == args.mode:
            pca_on_hidden_inference(args)
        elif 'find_learning_rate' == args.mode:
            _find_learning_rate(args)
        elif 'find_learning_rate_and_train' == args.mode:
            args.learning_rate = _find_learning_rate(args)
            if not args.learning_rate:
                raise ValueError('Could not find learning rate.')
            train_legacy(args)
        else:
            raise ValueError('Unknown mode:', args.mode)

    except Exception as e:
        logging.exception(e)

    if args.gcs_cloud_bucket is not None:
        save_to_google_cloud(args)

    end_time = timer()
    elapsed_time = end_time - start_time
    logging.info("Executed the '{}' operation in {:.2f} seconds".format(args.mode, elapsed_time))


def save_to_google_cloud(args):

    """
    Function to transfer all files and result from output folder to designated location google cloud bucket.
    Args:
        args.output_folder(str): Local folder path where all files and results generated from run are saved:
        args.gcs_cloud_bucket (str): Google cloud bucket folder path where all files will be stored
    """
    #Instantiate a storage client
    client = storage.Client()

    #get the bucket
    split_path = args.gcs_cloud_bucket.split("/")
    bucket_name = split_path[0]
    blob_path = "/".join(split_path[1:])

    bucket = client.get_bucket(bucket_name)

    # get the output folder
    if args.mode in ['train', 'compare', 'plot_predictions', 'infer_stats_from_segmented_regions']:
        output_folder = os.path.join(args.output_folder, args.id)
    else:
        output_folder = args.output_folder

    # uploading all files from local to server
    for root,_,files in os.walk(output_folder):
        for filename in files:
            local_file_path = os.path.join(root,filename)
            blob = bucket.blob(blob_path+filename)
            blob.upload_from_filename(local_file_path)
            print("Uploaded from local file path {} to {} in google cloud bucket".format(local_file_path,filename))





def _find_learning_rate(args) -> float:
    schedule = args.learning_rate_schedule
    args.learning_rate_schedule = None  # learning rate schedule interferes with setting lr done by find_learning_rate
    generate_train, _, _ = test_train_valid_tensor_generators(**args.__dict__)
    model = legacy_multimodal_multitask_model(**args.__dict__)
    lr = find_learning_rate(model, generate_train, args.training_steps, os.path.join(args.output_folder, args.id))
    args.learning_rate_schedule = schedule
    return lr


def train_legacy(args):
    generate_train, generate_valid, generate_test = test_train_valid_tensor_generators(**args.__dict__)
    model = legacy_multimodal_multitask_model(**args.__dict__)
    model = train_model_from_generators(
        model, generate_train, generate_valid, args.training_steps, args.validation_steps, args.batch_size, args.epochs,
        args.patience, args.output_folder, args.id, args.inspect_model, args.inspect_show_labels, args.tensor_maps_out,
        save_last_model=args.save_last_model,
    )

    out_path = os.path.join(args.output_folder, args.id + '/')
    test_data, test_labels, test_paths = big_batch_from_minibatch_generator(generate_test, args.test_steps)
    return _predict_and_evaluate(
        model, test_data, test_labels, args.tensor_maps_in, args.tensor_maps_out, args.tensor_maps_protected,
        args.batch_size, args.hidden_layer, out_path, test_paths, args.embed_visualization, args.alpha,
        args.dpi, args.plot_width, args.plot_height,
    )


def train_multimodal_multitask(args):
    generate_train, generate_valid, generate_test = test_train_valid_tensor_generators(**args.__dict__)
    model, encoders, decoders, merger = make_multimodal_multitask_model(**args.__dict__)
    model = train_model_from_generators(
        model, generate_train, generate_valid, args.training_steps, args.validation_steps, args.batch_size, args.epochs,
        args.patience, args.output_folder, args.id, args.inspect_model, args.inspect_show_labels, args.tensor_maps_out,
        save_last_model=args.save_last_model,
    )

    save_dir = os.path.join(args.output_folder, args.id)
    os.makedirs(save_dir, exist_ok=True)
    for tm in encoders:
        encoders[tm].save(os.path.join(save_dir, f'encoder_{tm.name}{MODEL_EXT}'))
    for tm in decoders:
        decoders[tm].save(os.path.join(save_dir, f'decoder_{tm.name}{MODEL_EXT}'))
    if merger:
        merger.save(os.path.join(save_dir, f'merger{MODEL_EXT}'))

    performance_metrics = {}
    if args.test_steps > 0:
        iter_generate_test = iter(generate_test)
        test_data, test_labels, _ = big_batch_from_minibatch_generator(iter_generate_test, args.test_steps)
        test_paths = None
        performance_metrics = _predict_and_evaluate(
            model, test_data, test_labels, args.tensor_maps_in, args.tensor_maps_out, args.tensor_maps_protected,
            args.batch_size, args.hidden_layer, os.path.join(args.output_folder, args.id + '/'), test_paths,
            args.embed_visualization, args.alpha, args.dpi, args.plot_width, args.plot_height,
        )

        predictions = model.predict(test_data)
        predictions_dict = {}

        if isinstance(predictions, dict):
            predictions_dict = predictions

        elif isinstance(predictions, (list, tuple)):
            # Map outputs by model.output_names (strings)
            predictions_dict = {
                str(name): pred for name, pred in zip(model.output_names, predictions)
            }
        else:
            # Single tensor output
            predictions_dict = {
                model.output_names[0]: predictions
        }
        samples = min(args.test_steps * args.batch_size, 12)
        out_path = os.path.join(args.output_folder, args.id, 'reconstructions/')
        logging.info(f'Predictions and shapes are: {[(p, predictions_dict[p].shape) for p in predictions_dict]}')

        for i, etm in enumerate(encoders):
            embed = encoders[etm].predict(test_data[etm.input_name()])
            if etm.output_name() in predictions_dict:
                plot_reconstruction(etm, test_data[etm.input_name()], predictions_dict[etm.output_name()], out_path, test_paths, samples)
            for dtm in decoders:
                reconstruction = decoders[dtm].predict(embed)
                logging.info(f'{dtm.name} has prediction shape: {reconstruction.shape} from embed shape: {embed.shape}')
                my_out_path = os.path.join(out_path, f'decoding_{dtm.name}_from_{etm.name}/')
                os.makedirs(os.path.dirname(my_out_path), exist_ok=True)
                if dtm.axes() > 1:
                    plot_reconstruction(dtm, test_labels[dtm.output_name()], reconstruction, my_out_path, test_paths, samples)
                else:
                    evaluate_predictions(
                        dtm, reconstruction, test_labels[dtm.output_name()], {}, dtm.name, my_out_path,
                        test_paths, dpi=args.dpi, width=args.plot_width, height=args.plot_height,
                    )
    return performance_metrics


def test_multimodal_multitask(args):
    _, _, generate_test = test_train_valid_tensor_generators(**args.__dict__)
    model, _, _, _ = make_multimodal_multitask_model(**args.__dict__)
    out_path = os.path.join(args.output_folder, args.id + '/')
    data, labels, paths = big_batch_from_minibatch_generator(generate_test, args.test_steps)
    return _predict_and_evaluate(
        model, data, labels, args.tensor_maps_in, args.tensor_maps_out, args.tensor_maps_protected, args.batch_size,
        args.hidden_layer, out_path, paths, args.embed_visualization, args.alpha,
        args.dpi, args.plot_width, args.plot_height,
    )


def test_multimodal_scalar_tasks(args):
    _, _, generate_test = test_train_valid_tensor_generators(**args.__dict__)
    model, _, _, _ = make_multimodal_multitask_model(**args.__dict__)
    p = os.path.join(args.output_folder, args.id + '/')
    return _predict_scalars_and_evaluate_from_generator(
        model, generate_test, args.tensor_maps_in, args.tensor_maps_out, args.tensor_maps_protected, args.test_steps,
        args.hidden_layer, p, args.alpha, args.dpi, args.plot_width, args.plot_height,
    )


def compare_multimodal_multitask_models(args):
    _, _, generate_test = test_train_valid_tensor_generators(**args.__dict__)
    models_inputs_outputs = get_model_inputs_outputs(args.model_files, args.tensor_maps_in, args.tensor_maps_out)
    input_data, output_data, paths = big_batch_from_minibatch_generator(generate_test, args.test_steps)
    common_outputs = _get_common_outputs(models_inputs_outputs, 'output')
    predictions = _get_predictions(args, models_inputs_outputs, input_data, common_outputs, 'input', 'output')
    _calculate_and_plot_prediction_stats(args, predictions, output_data, paths)


def compare_multimodal_scalar_task_models(args):
    _, _, generate_test = test_train_valid_tensor_generators(**args.__dict__)
    models_io = get_model_inputs_outputs(args.model_files, args.tensor_maps_in, args.tensor_maps_out)
    outs = _get_common_outputs(models_io, "output")
    predictions, labels, paths = _scalar_predictions_from_generator(args, models_io, generate_test, args.test_steps, outs, "input", "output")
    _calculate_and_plot_prediction_stats(args, predictions, labels, paths)


def standardize_by_sample_ecg(ecg, _):
    """Z Score ECG"""
    return (ecg - np.mean(ecg)) / (np.std(ecg) + 1e-6)


def train_xdl(args):
    mrn_df = pd.read_csv(args.app_csv)
    if 'start_fu_age' in mrn_df:
        mrn_df['age_in_days'] = pd.to_timedelta(mrn_df.start_fu_age).dt.days
    elif 'start_fu' in mrn_df:
        mrn_df['age_in_days'] = pd.to_timedelta(mrn_df.start_fu).dt.days
    for ot in args.tensor_maps_out:
        mrn_df = mrn_df[mrn_df[ot.name].notna()]
    mrn_df = mrn_df.set_index('sample_id')

    output_dds = [dataframe_data_description_from_tensor_map(tmap, mrn_df) for tmap in args.tensor_maps_out]

    ecg_dd = ECGDataDescription(
        args.tensors,
        name=args.tensor_maps_in[0].input_name(),
        ecg_len=5000,  # all ECGs will be linearly interpolated to be this length
        transforms=[standardize_by_sample_ecg],  # these will be applied in order
        # data will be automatically localized from s3
    )

    def option_picker(sample_id, data_descriptions):
        ecg_dts = ecg_dd.get_loading_options(sample_id)
        htn_dt = output_dds[0].get_loading_options(sample_id)[0]['start_fu_datetime']
        min_ecg_dt = htn_dt - pd.to_timedelta("1095d")
        max_ecg_dt = htn_dt + pd.to_timedelta("1095d")
        dates = []
        for dt in ecg_dts:
            if min_ecg_dt <= dt[DATE_OPTION_KEY] <= max_ecg_dt:
                dates.append(dt)
        if len(dates) == 0:
            raise ValueError('No matching dates')
        chosen_dt = np.random.choice(dates)
        chosen_dt['day_delta'] = (htn_dt - chosen_dt[DATE_OPTION_KEY]).days
        return {dd: chosen_dt for dd in data_descriptions}

    sg = DataDescriptionSampleGetter(
        input_data_descriptions=[ecg_dd],  # what we want a model to use as input data
        output_data_descriptions=output_dds,  # what we want a model to predict from the input data
        option_picker=option_picker,
    )
    model, encoders, decoders, merger = make_multimodal_multitask_model(**args.__dict__)

    train_ids = list(mrn_df[mrn_df.split == 'train'].index)
    valid_ids = list(mrn_df[mrn_df.split == 'valid'].index)
    test_ids = list(mrn_df[mrn_df.split == 'test'].index)

    train_dataset = SampleGetterIterableDataset(
        sample_ids=list(train_ids), sample_getter=sg,
        get_epoch=shuffle_get_epoch,
    )
    valid_dataset = SampleGetterIterableDataset(
        sample_ids=list(valid_ids), sample_getter=sg,
        get_epoch=shuffle_get_epoch,
    )

    num_train_workers = int(args.training_steps / (args.training_steps + args.validation_steps) * args.num_workers) or (1 if args.num_workers else 0)
    num_valid_workers = int(args.validation_steps / (args.training_steps + args.validation_steps) * args.num_workers) or (1 if args.num_workers else 0)

    generate_train = TensorMapDataLoader2(
        batch_size=args.batch_size, input_maps=args.tensor_maps_in, output_maps=args.tensor_maps_out,
        dataset=train_dataset,
        num_workers=num_train_workers,
    )
    generate_valid = TensorMapDataLoader2(
        batch_size=args.batch_size, input_maps=args.tensor_maps_in, output_maps=args.tensor_maps_out,
        dataset=valid_dataset,
        num_workers=num_valid_workers,
    )

    model = train_model_from_generators(
        model, generate_train, generate_valid, args.training_steps, args.validation_steps, args.batch_size, args.epochs,
        args.patience, args.output_folder, args.id, args.inspect_model, args.inspect_show_labels, args.tensor_maps_out,
        save_last_model=args.save_last_model,
    )
    for tm in encoders:
        encoders[tm].save(f'{args.output_folder}{args.id}/encoder_{tm.name}.h5')
    for tm in decoders:
        decoders[tm].save(f'{args.output_folder}{args.id}/decoder_{tm.name}.h5')
    if merger:
        merger.save(f'{args.output_folder}{args.id}/merger.h5')


def train_xdl_af(args):
    mrn_df = pd.read_csv(args.app_csv)
    mrn_df = mrn_df.dropna(subset=['last_encounter'])
    mrn_df.MRN = mrn_df.MRN.astype(int)
    mrn_df['survival_curve_af'] = mrn_df.af_event
    #mrn_df['start_date'] = mrn_df.start_fu_datetime

    if 'start_fu_age' in mrn_df:
        mrn_df['age_in_days'] = pd.to_timedelta(mrn_df.start_fu_age).dt.days
    elif 'start_fu' in mrn_df:
        mrn_df['age_in_days'] = pd.to_timedelta(mrn_df.start_fu).dt.days
    for ot in args.tensor_maps_out:
        mrn_df = mrn_df[mrn_df[ot.name].notna()]

    mrn_df = mrn_df.set_index('MRN')

    output_dds = [dataframe_data_description_from_tensor_map(tmap, mrn_df) for tmap in args.tensor_maps_out]

    # ecg_dd = ECGDataDescription(
    #         args.tensors,
    #         name=f'input_ecg_strip_I_continuous',
    #         ecg_len=5000,  # all ECGs will be linearly interpolated to be this length
    #         transforms=[standardize_by_sample_ecg],  # these will be applied in order
    #         leads={'I': 0},
    #     )
    ecg_dd = ECGDataDescription(
        args.tensors,
        name=args.tensor_maps_in[0].input_name(),
        ecg_len=5000,  # all ECGs will be linearly interpolated to be this length
        transforms=[standardize_by_sample_ecg],  # these will be applied in order
        # data will be automatically localized from s3
    )

    def option_picker(sample_id, data_descriptions):
        ecg_dts = ecg_dd.get_loading_options(sample_id)
        start_dt = output_dds[0].get_loading_options(sample_id)[0]['start_date']
        min_ecg_dt = start_dt - pd.to_timedelta("1095d")
        dates = []
        for dt in ecg_dts:
            if min_ecg_dt <= dt[DATE_OPTION_KEY] <= start_dt:
                dates.append(dt)
        if len(dates) == 0:
            raise ValueError('No matching dates')
        chosen_dt = np.random.choice(dates)
        chosen_dt['start_date'] = start_dt
        chosen_dt['day_delta'] = (start_dt - chosen_dt[DATE_OPTION_KEY]).days
        return {dd: chosen_dt for dd in data_descriptions}

    logging.info(f'output_dds[0].name {output_dds[0].name}')
    logging.info(f'output_dds[0].get_loading_options(sample_id)[0] {output_dds[0].get_loading_options(3773)[0]}')
    logging.info(f'option_picker {option_picker(3773, [ecg_dd])}')

    sg = DataDescriptionSampleGetter(
        input_data_descriptions=[ecg_dd],  # what we want a model to use as input data
        output_data_descriptions=output_dds,  # what we want a model to predict from the input data
        option_picker=option_picker,
    )

    b = sg(3773)
    logging.info(f'batch output: {b[1]}')

    model, encoders, decoders, merger = make_multimodal_multitask_model(**args.__dict__)

    train_ids = list(mrn_df[mrn_df.split == 'train'].index)
    valid_ids = list(mrn_df[mrn_df.split == 'valid'].index)
    test_ids = list(mrn_df[mrn_df.split == 'test'].index)

    train_dataset = SampleGetterIterableDataset(
        sample_ids=list(train_ids), sample_getter=sg,
        get_epoch=shuffle_get_epoch,
    )
    valid_dataset = SampleGetterIterableDataset(
        sample_ids=list(valid_ids), sample_getter=sg,
        get_epoch=shuffle_get_epoch,
    )

    num_train_workers = int(args.training_steps / (args.training_steps + args.validation_steps) * args.num_workers) or (1 if args.num_workers else 0)
    num_valid_workers = int(args.validation_steps / (args.training_steps + args.validation_steps) * args.num_workers) or (1 if args.num_workers else 0)

    generate_train = TensorMapDataLoader2(
        batch_size=args.batch_size, input_maps=args.tensor_maps_in, output_maps=args.tensor_maps_out,
        dataset=train_dataset,
        num_workers=num_train_workers,
    )
    generate_valid = TensorMapDataLoader2(
        batch_size=args.batch_size, input_maps=args.tensor_maps_in, output_maps=args.tensor_maps_out,
        dataset=valid_dataset,
        num_workers=num_valid_workers,
    )

    model = train_model_from_generators(
        model, generate_train, generate_valid, args.training_steps, args.validation_steps, args.batch_size, args.epochs,
        args.patience, args.output_folder, args.id, args.inspect_model, args.inspect_show_labels, args.tensor_maps_out,
        save_last_model=args.save_last_model,
    )
    for tm in encoders:
        encoders[tm].save(f'{args.output_folder}{args.id}/encoder_{tm.name}.h5')
    for tm in decoders:
        decoders[tm].save(f'{args.output_folder}{args.id}/decoder_{tm.name}.h5')
    if merger:
        merger.save(f'{args.output_folder}{args.id}/merger.h5')

    test_sg = DataDescriptionSampleGetter(
        input_data_descriptions=[ecg_dd],  # what we want a model to use as input data
        output_data_descriptions=output_dds,  # what we want a model to predict from the input data
        option_picker=option_picker,
    )
    test_dataset = SampleGetterIterableDataset(
        sample_ids=list(test_ids), sample_getter=test_sg,
        get_epoch=shuffle_get_epoch,
    )

    generate_test = TensorMapDataLoader2(
        batch_size=args.batch_size, input_maps=args.tensor_maps_in, output_maps=args.tensor_maps_out,
        dataset=test_dataset,
        num_workers=num_train_workers,
    )

    y_trues = defaultdict(list)
    y_preds = defaultdict(list)
    logging.info(f'Now start testing... output keys: {list(next(generate_test)[1].keys())}')
    for X, y in generate_test:
        preds = model.predict(X, verbose=0)
        if len(model.output_names) == 1:
            preds = [preds]
        predictions_dict = {name: pred for name, pred in zip(model.output_names, preds)}
        for otm in args.tensor_maps_out:
            y_preds[otm.name].extend(predictions_dict[otm.output_name()])
            y_trues[otm.name].extend(y[otm.output_name()])

        if len(y_trues[otm.name]) % 400 == 0:
            print(f'Predicted on {len(y_trues[otm.name])} test examples.')
        if len(y_trues[otm.name]) > args.test_steps*args.batch_size:
            break

    for otm in args.tensor_maps_out:
        y_preds[otm.name] = np.array(y_preds[otm.name])
        y_trues[otm.name] = np.array(y_trues[otm.name])

    for otm in args.tensor_maps_out:
        if otm.is_survival_curve():
            plot_survival(y_preds[otm.name], y_trues[otm.name], f'{otm.name.upper()} Model:{args.id}', otm.days_window)
        elif otm.is_categorical():
            plot_roc(y_preds[otm.name], y_trues[otm.name], otm.channel_map, f'{otm.name} ROC')
            plot_precision_recall_per_class(
                y_preds[otm.name], y_trues[otm.name], otm.channel_map,
                f'{otm.name} Precision Recall',
            )
        elif otm.is_continuous():
            plot_scatter(y_preds[otm.name], y_trues[otm.name], f'{otm.name} Scatter')


def datetime_to_float(d):
    return pd.to_datetime(d, utc=True).timestamp()


def float_to_datetime(fl):
    return pd.to_datetime(fl, unit='s', utc=True)


def infer_from_dataloader(dataloader, model, tensor_maps_out, max_batches=125000):
    dataloader_iterator = iter(dataloader)
    space_dict = defaultdict(list)
    for i in range(max_batches):
        try:
            data, target = next(dataloader_iterator)
            for k in data:
                data[k] = np.array(data[k])
            prediction = model.predict(data, verbose=0)
            if len(model.output_names) == 1:
                prediction = [prediction]
            predictions_dict = {name: pred for name, pred in zip(model.output_names, prediction)}
            for b in range(prediction[0].shape[0]):
                for otm in tensor_maps_out:
                    y = predictions_dict[otm.output_name()]
                    if otm.is_categorical():
                        space_dict[f'{otm.name}_prediction'].append(y[b, 1])
                    elif otm.is_continuous():
                        space_dict[f'{otm.name}_prediction'].append(y[b, 0])
                    elif otm.is_survival_curve():
                        intervals = otm.shape[-1] // 2
                        days_per_bin = 1 + (2*otm.days_window) // intervals
                        predicted_survivals = np.cumprod(y[:, :intervals], axis=1)
                        space_dict[f'{otm.name}_prediction'].append(str(1 - predicted_survivals[0, -1]))
                        sick = np.sum(target[otm.output_name()][:, intervals:], axis=-1)
                        follow_up = np.cumsum(target[otm.output_name()][:, :intervals], axis=-1)[:, -1] * days_per_bin
                        space_dict[f'{otm.name}_event'].append(str(sick[0]))
                        space_dict[f'{otm.name}_follow_up'].append(str(follow_up[0]))
                for k in target:
                    if k in ['MRN', 'linker_id', 'is_c3po', 'output_age_in_days_continuous']:
                        space_dict[f'{k}'].append(target[k][b].numpy())
                    elif k in ['datetime']:
                        space_dict[f'{k}'].append(float_to_datetime(int(target[k][b].numpy())))
                    else:
                        space_dict[f'{k}'].append(target[k][b, -1].numpy())
            if i % 100 == 0:
                logging.info(f'Inferred on {i} batches, {len(space_dict[k])} rows')
        except StopIteration:
            logging.info(f'Inferred on all {i} batches.')
            break
    return pd.DataFrame.from_dict(space_dict)


def infer_xdl(args):
    mrn_df = pd.read_csv(args.app_csv)
    if 'start_fu_age' in mrn_df:
        mrn_df['age_in_days'] = pd.to_timedelta(mrn_df.start_fu_age).dt.days
    elif 'start_fu' in mrn_df:
        mrn_df['age_in_days'] = pd.to_timedelta(mrn_df.start_fu).dt.days
    mrn_df = mrn_df.rename(columns={'Dem.Gender.no_filter_x': 'sex', 'Dem.Gender.no_filter': 'sex'})
    mrn_df['is_c3po'] = mrn_df.cohort == 'c3po'
    for ot in args.tensor_maps_out:
        mrn_df = mrn_df[mrn_df[ot.name].notna()]
    mrn_df = mrn_df.set_index('sample_id')

    output_dds = [dataframe_data_description_from_tensor_map(tmap, mrn_df) for tmap in args.tensor_maps_out]

    output_dds.append(DataFrameDataDescription(mrn_df, col="MRN"))
    output_dds.append(DataFrameDataDescription(mrn_df, col="linker_id"))
    output_dds.append(DataFrameDataDescription(mrn_df, col="is_c3po"))
    output_dds.append(DataFrameDataDescription(mrn_df, col="datetime", process_col=datetime_to_float))

    ecg_dd = ECGDataDescription(
        args.tensors,
        name=args.tensor_maps_in[0].input_name(),
        ecg_len=5000,  # all ECGs will be linearly interpolated to be this length
        transforms=[standardize_by_sample_ecg],  # these will be applied in order
    )

    def test_option_picker(sample_id, data_descriptions):
        ecg_dts = ecg_dd.get_loading_options(sample_id)
        htn_dt = output_dds[0].get_loading_options(sample_id)[0]['start_fu_datetime']
        min_ecg_dt = htn_dt - pd.to_timedelta("1095d")
        max_ecg_dt = htn_dt - pd.to_timedelta("1d")
        dates = []
        for dt in ecg_dts:
            if min_ecg_dt <= dt[DATE_OPTION_KEY] <= max_ecg_dt:
                dates.append(dt)
        if len(dates) == 0:
            raise ValueError('No matching dates')
        chosen_dt = dates[-1]
        return {dd: chosen_dt for dd in data_descriptions}

    sg = DataDescriptionSampleGetter(
        input_data_descriptions=[ecg_dd],  # what we want a model to use as input data
        output_data_descriptions=output_dds,  # what we want a model to predict from the input data
        option_picker=test_option_picker,
    )

    dataset = SampleGetterIterableDataset(sample_ids=list(mrn_df.index), sample_getter=sg, get_epoch=shuffle_get_epoch)
    dataloader = DataLoader(dataset, num_workers=0, batch_size=args.batch_size)

    model, _, _, _ = make_multimodal_multitask_model(**args.__dict__)
    infer_df = infer_from_dataloader(dataloader, model, args.tensor_maps_out)
    if 'mgh' in args.tensors:
        hospital = 'mgh'
        infer_df = infer_df.rename(columns={'MRN': 'MGH_MRN'})
        infer_df.MGH_MRN = infer_df.MGH_MRN.astype(int)
    else:
        hospital = 'bwh'
        infer_df = infer_df.rename(columns={'MRN': 'BWH_MRN'})
        infer_df.BWH_MRN = infer_df.BWH_MRN.astype(int)

    infer_df.linker_id = infer_df.linker_id.astype(int)
    names = '_'.join([otm.name for otm in args.tensor_maps_out])
    now_string = datetime.datetime.now().strftime('%Y_%m_%d')
    out_file = f'{args.output_folder}/{args.id}/infer_{names}_{hospital}_v{now_string}.tsv'
    infer_df.to_csv(out_file, sep='\t', index=False)
    logging.info(f'Infer dataframe head: {infer_df.head()}  \n\n Saved inferences to: {out_file}')


def _make_tmap_nan_on_fail(tmap):
    """
    Builds a copy TensorMap with a tensor_from_file that returns nans on errors instead of raising an error
    """
    new_tmap = copy.deepcopy(tmap)
    new_tmap.validator = lambda _, x, hd5: x  # prevent failure caused by validator

    def _tff(tm, hd5, dependents=None):
        try:
            return tmap.tensor_from_file(tm, hd5, dependents)
        except (IndexError, KeyError, ValueError, OSError, RuntimeError):
            return np.full(shape=tm.shape, fill_value=np.nan)

    new_tmap.tensor_from_file = _tff
    return new_tmap


def inference_file_name(output_folder: str, id_: str) -> str:
    return os.path.join(output_folder, id_, 'inference_' + id_ + '.tsv')


def infer_multimodal_multitask(args):
    stats = Counter()
    tensor_paths_inferred = set()
    inference_tsv = inference_file_name(args.output_folder, args.id)
    tsv_style_is_genetics = 'genetics' in args.tsv_style

    model = legacy_multimodal_multitask_model(**args.__dict__)
    no_fail_tmaps_out = [_make_tmap_nan_on_fail(tmap) for tmap in args.tensor_maps_out]
    tensor_paths = _tensor_paths_from_sample_csv(args.tensors, args.sample_csv)
    # hard code batch size to 1 so we can iterate over file names and generated tensors together in the tensor_paths for loop
    generate_test = TensorGenerator(
        1, args.tensor_maps_in, no_fail_tmaps_out, tensor_paths, num_workers=0,
        cache_size=0, keep_paths=True, mixup_alpha=args.mixup_alpha,
    )

    logging.info(f"Found {len(tensor_paths)} tensor paths.")
    generate_test.set_worker_paths(tensor_paths)
    output_maps = {tm.output_name(): tm for tm in no_fail_tmaps_out}
    with open(inference_tsv, mode='w') as inference_file:
        # TODO: csv.DictWriter is much nicer for this
        inference_writer = csv.writer(inference_file, delimiter='\t', quotechar='"', quoting=csv.QUOTE_MINIMAL)
        header = ['sample_id']
        if tsv_style_is_genetics:
            header = ['FID', 'IID']
        for ot in model.output_names:
            otm = output_maps[ot]
            logging.info(f"Got ot  {ot} and otm {otm}  ot and otm {otm.name} ot  and otm {otm.channel_map} channel_map and otm {otm.interpretation}.")
            if len(otm.shape) == 1 and otm.is_continuous():
                header.extend([otm.name + '_prediction', otm.name + '_actual'])
            elif len(otm.shape) == 1 and otm.is_categorical():
                channel_columns = []
                for k in otm.channel_map:
                    channel_columns.append(otm.name + '_' + k + '_prediction')
                    channel_columns.append(otm.name + '_' + k + '_actual')
                header.extend(channel_columns)
            elif otm.is_survival_curve():
                header.extend([otm.name + '_prediction', otm.name + '_actual', otm.name + '_follow_up'])
        inference_writer.writerow(header)

        while True:
            batch = next(generate_test)
            input_data, output_data, tensor_paths = batch[BATCH_INPUT_INDEX], batch[BATCH_OUTPUT_INDEX], batch[BATCH_PATHS_INDEX]
            if tensor_paths[0] in tensor_paths_inferred:
                next(generate_test)  # this prints end of epoch info
                logging.info(f"Inference on {stats['count']} tensors finished. Inference TSV file at: {inference_tsv}")
                break
            prediction = model.predict(input_data, verbose=0)
            if len(no_fail_tmaps_out) == 1:
                prediction = [prediction]
            predictions_dict = {name: pred for name, pred in zip(model.output_names, prediction)}
            sample_id = os.path.basename(tensor_paths[0]).replace(TENSOR_EXT, '')
            csv_row = [sample_id]
            if tsv_style_is_genetics:
                csv_row *= 2
            for ot in model.output_names:
                y = predictions_dict[ot]
                otm = output_maps[ot]
                if len(otm.shape) == 1 and otm.is_continuous():
                    csv_row.append(str(otm.rescale(y)[0][0]))  # first index into batch then index into the 1x1 structure
                    if ((otm.sentinel is not None and otm.sentinel == output_data[otm.output_name()][0][0])
                            or np.isnan(output_data[otm.output_name()][0][0])):
                        csv_row.append("NA")
                    else:
                        csv_row.append(str(otm.rescale(output_data[otm.output_name()])[0][0]))
                elif len(otm.shape) == 1 and otm.is_categorical():
                    for k, i in otm.channel_map.items():
                        try:
                            csv_row.append(str(y[ot][0][otm.channel_map[k]]))
                            actual = output_data[otm.output_name()][0][i]
                            csv_row.append("NA" if np.isnan(actual) else str(actual))
                        except IndexError:
                            logging.debug(f'index error at {otm.name} item {i} key {k} with cm: {otm.channel_map} y is {y.shape} y is {y}')
                elif otm.is_survival_curve():
                    intervals = otm.shape[-1] // 2
                    days_per_bin = 1 + otm.days_window // intervals
                    #predicted_survivals = np.cumprod(y[:, :intervals], axis=1)
                    predicted_survivals = np.cumprod(y[:, :10], axis=1)  # 2 year probability
                    csv_row.append(str(1 - predicted_survivals[0, -1]))
                    sick = np.sum(output_data[otm.output_name()][:, intervals:], axis=-1)
                    follow_up = np.cumsum(output_data[otm.output_name()][:, :intervals], axis=-1)[:, -1] * days_per_bin
                    csv_row.extend([str(sick[0]), str(follow_up[0])])
                elif otm.axes() > 1:
                    hd5_path = os.path.join(args.output_folder, args.id, 'inferred_hd5s', f'{sample_id}{TENSOR_EXT}')
                    os.makedirs(os.path.dirname(hd5_path), exist_ok=True)
                    with h5py.File(hd5_path, 'a') as hd5:
                        hd5.create_dataset(
                            f'{otm.name}_truth', data=otm.rescale(output_data[otm.output_name()][0]),
                            compression='gzip',
                        )
                        if otm.path_prefix == 'ukb_ecg_rest':
                            for lead in otm.channel_map:
                                hd5.create_dataset(
                                    f'/ukb_ecg_rest/{lead}/instance_0',
                                    data=otm.rescale(y[0, otm.channel_map[lead]]),
                                    compression='gzip',
                                )
            inference_writer.writerow(csv_row)
            tensor_paths_inferred.add(tensor_paths[0])
            stats['count'] += 1
            if stats['count'] % 250 == 0:
                logging.info(f"Wrote:{stats['count']} rows of inference.  Last tensor:{tensor_paths[0]}")


def _tensor_paths_from_sample_csv(tensors, sample_csv):
    sample_set = None
    if sample_csv is not None:
        with open(sample_csv, 'r') as csv_file:
            sample_ids = [row[0] for row in csv.reader(csv_file)]
            sample_set = set(sample_ids[1:])
    tensor_paths = [
        file for file in glob.glob(os.path.join(tensors, f"*{TENSOR_EXT}"))
        if sample_set is None or os.path.splitext(os.path.basename(file))[0] in sample_set
    ]
    return tensor_paths


def _hidden_file_name(output_folder: str, prefix_: str, id_: str, extension_: str) -> str:
    return os.path.join(output_folder, id_, f'hidden_{prefix_}_{id_}{extension_}')


def infer_hidden_layer_multimodal_multitask(args):
    stats = Counter()
    args.num_workers = 0
    inference_tsv = _hidden_file_name(args.output_folder, args.hidden_layer, args.id, '.tsv')
    tsv_style_is_genetics = 'genetics' in args.tsv_style
    tensor_paths = _tensor_paths_from_sample_csv(args.tensors, args.sample_csv)
    # hard code batch size to 1 so we can iterate over file names and generated tensors together in the tensor_paths for loop
    generate_test = TensorGenerator(
        1, args.tensor_maps_in, args.tensor_maps_out, tensor_paths, num_workers=0,
        cache_size=args.cache_size, keep_paths=True, mixup_alpha=args.mixup_alpha,
    )
    generate_test.set_worker_paths(tensor_paths)
    full_model = legacy_multimodal_multitask_model(**args.__dict__)
    embed_model = make_hidden_layer_model(full_model, args.tensor_maps_in, args.hidden_layer)
    embed_model.save(_hidden_file_name(args.output_folder, f'{args.hidden_layer}_encoder_', args.id, '.h5'))
    dummy_input = {tm.input_name(): np.zeros((1,) + tuple(tm.shape)) for tm in args.tensor_maps_in}
    dummy_out = embed_model.predict(dummy_input)
    latent_dimensions = int(np.prod(dummy_out.shape[1:]))
    logging.info(f'Dummy output shape is: {dummy_out.shape} latent dimensions: {latent_dimensions} Will write inferences to: {inference_tsv}')
    with open(inference_tsv, mode='w') as inference_file:
        inference_writer = csv.writer(inference_file, delimiter='\t', quotechar='"', quoting=csv.QUOTE_MINIMAL)
        header = ['FID', 'IID'] if tsv_style_is_genetics else ['sample_id']
        header += [f'latent_{i}' for i in range(latent_dimensions)]
        inference_writer.writerow(header)

        while True:
            batch = next(generate_test)
            input_data, tensor_paths = batch[BATCH_INPUT_INDEX], batch[BATCH_PATHS_INDEX]
            if tensor_paths[0] in stats:
                next(generate_test)  # this prints end of epoch info
                logging.info(f"Latent space inference on {stats['count']} tensors finished. Inference TSV file at: {inference_tsv}")
                break

            sample_id = os.path.basename(tensor_paths[0]).replace(TENSOR_EXT, '')
            prediction = embed_model.predict(input_data, verbose=0)
            prediction = np.reshape(prediction, (latent_dimensions,))
            csv_row = [sample_id, sample_id] if tsv_style_is_genetics else [sample_id]
            csv_row += [f'{prediction[i]}' for i in range(latent_dimensions)]
            inference_writer.writerow(csv_row)
            stats[tensor_paths[0]] += 1
            stats['count'] += 1
            if stats['count'] % 500 == 0:
                logging.info(f"Wrote:{stats['count']} rows of latent space inference.  Last tensor:{tensor_paths[0]}")


def train_shallow_model(args):
    generate_train, generate_valid, generate_test = test_train_valid_tensor_generators(**args.__dict__)
    model = make_shallow_model(args.tensor_maps_in, args.tensor_maps_out, args.learning_rate, args.model_file, args.model_layers)
    model = train_model_from_generators(
        model, generate_train, generate_valid, args.training_steps, args.validation_steps, args.batch_size,
        args.epochs, args.patience, args.output_folder, args.id, args.inspect_model, args.inspect_show_labels,
    )

    p = os.path.join(args.output_folder, args.id + '/')
    test_data, test_labels, test_paths = big_batch_from_minibatch_generator(generate_test, args.test_steps)
    return _predict_and_evaluate(
        model, test_data, test_labels, args.tensor_maps_in, args.tensor_maps_out, args.tensor_maps_protected,
        args.batch_size, args.hidden_layer, p, test_paths, args.embed_visualization, args.alpha,
        args.dpi, args.plot_width, args.plot_height,
    )


def train_siamese_model(args):
    base_model = legacy_multimodal_multitask_model(**args.__dict__)
    siamese_model = make_siamese_model(base_model, **args.__dict__)
    generate_train, generate_valid, generate_test = test_train_valid_tensor_generators(**args.__dict__, siamese=True)
    siamese_model = train_model_from_generators(
        siamese_model, generate_train, generate_valid, args.training_steps, args.validation_steps, args.batch_size,
        args.epochs, args.patience, args.output_folder, args.id, args.inspect_model, args.inspect_show_labels,
    )

    data, labels, paths = big_batch_from_minibatch_generator(generate_test, args.test_steps)
    prediction = siamese_model.predict(data)
    return subplot_roc_per_class(
        prediction, labels['output_siamese'], {'random_siamese_verification_task': 0},
        args.protected_maps, args.id, os.path.join(args.output_folder, args.id + '/'),
    )


def infer_encoders_block_multimodal_multitask(args):
    args.num_workers = 0
    tsv_style_is_genetics = 'genetics' in args.tsv_style
    sample_set = None
    if args.sample_csv is not None:
        with open(args.sample_csv, 'r') as csv_file:
            sample_ids = [row[0] for row in csv.reader(csv_file)]
            sample_set = set(sample_ids[1:])
    _, encoders, _, _ = make_multimodal_multitask_model(**args.__dict__)
    latent_dimensions = args.dense_layers[-1]
    for e in encoders:
        stats = Counter()
        inference_tsv = _hidden_file_name(args.output_folder, e.name, args.id, '.tsv')
        logging.info(f'Will write encodings from {e.name} to: {inference_tsv}')
        # hard code batch size to 1 so we can iterate over file names and generated tensors together in the tensor_paths for loop
        tensor_paths = [
            os.path.join(args.tensors, tp) for tp in sorted(os.listdir(args.tensors))
            if os.path.splitext(tp)[-1].lower() == TENSOR_EXT and (sample_set is None or os.path.splitext(tp)[0] in sample_set)
        ]
        generate_test = TensorGenerator(
            1, [e], [], tensor_paths, num_workers=0,
            cache_size=args.cache_size, keep_paths=True, mixup_alpha=args.mixup_alpha,
        )
        generate_test.set_worker_paths(tensor_paths)
        with open(inference_tsv, mode='w') as inference_file:
            inference_writer = csv.writer(inference_file, delimiter='\t', quotechar='"', quoting=csv.QUOTE_MINIMAL)
            header = ['FID', 'IID'] if tsv_style_is_genetics else ['sample_id']
            header += [f'latent_{i}' for i in range(latent_dimensions)]
            inference_writer.writerow(header)

            while True:
                batch = next(generate_test)
                input_data, tensor_paths = batch[BATCH_INPUT_INDEX], batch[BATCH_PATHS_INDEX]
                if tensor_paths[0] in stats:
                    next(generate_test)  # this prints end of epoch info
                    logging.info(f"Latent space inference on {stats['count']} tensors finished. Inference TSV file at: {inference_tsv}")
                    del stats
                    break

                sample_id = os.path.basename(tensor_paths[0]).replace(TENSOR_EXT, '')
                prediction = encoders[e].predict(input_data, verbose=0)
                prediction = np.reshape(prediction, (latent_dimensions,))
                csv_row = [sample_id, sample_id] if tsv_style_is_genetics else [sample_id]
                csv_row += [f'{prediction[i]}' for i in range(latent_dimensions)]
                inference_writer.writerow(csv_row)
                stats[tensor_paths[0]] += 1
                stats['count'] += 1
                if stats['count'] % 500 == 0:
                    logging.info(f"Wrote:{stats['count']} rows of latent space inference.  Last tensor:{tensor_paths[0]}")


def pca_on_hidden_inference(args):
    latent_cols = [f'latent_{i}' for i in range(args.dense_layers[0])]
    pca_on_tsv(args.app_csv, latent_cols, 'sample_id', args.dense_layers[1])


def plot_predictions(args):
    _, _, generate_test = test_train_valid_tensor_generators(**args.__dict__)
    model, _, _, _ = make_multimodal_multitask_model(**args.__dict__)
    data, labels, paths = big_batch_from_minibatch_generator(generate_test, args.test_steps)
    predictions = model.predict(data, batch_size=args.batch_size)
    if len(args.tensor_maps_out) == 1:
        predictions = [predictions]
    folder = os.path.join(args.output_folder, args.id, 'prediction_pngs/')
    predictions_to_pngs(predictions, args.tensor_maps_in, args.tensor_maps_out, data, labels, paths, folder)


def plot_while_training(args):
    generate_train, _, generate_test = test_train_valid_tensor_generators(**args.__dict__)
    test_data, test_labels, test_paths = big_batch_from_minibatch_generator(generate_test, args.test_steps)
    model = legacy_multimodal_multitask_model(**args.__dict__)

    plot_folder = os.path.join(args.output_folder, args.id, 'training_frames/')
    plot_while_learning(
        model, args.tensor_maps_in, args.tensor_maps_out, generate_train, test_data, test_labels, test_paths, args.epochs,
        args.batch_size, args.training_steps, plot_folder, args.write_pngs,
    )


def saliency_maps(args):
    import tensorflow as tf
    tf.compat.v1.disable_eager_execution()
    from tensorflow.python.framework.ops import disable_eager_execution
    disable_eager_execution()
    _, _, generate_test = test_train_valid_tensor_generators(**args.__dict__)
    model = legacy_multimodal_multitask_model(**args.__dict__)
    data, labels, paths = big_batch_from_minibatch_generator(generate_test, args.test_steps)
    in_tensor = data[args.tensor_maps_in[0].input_name()]
    for tm in args.tensor_maps_out:
        if len(tm.shape) > 1:
            continue
        for channel in tm.channel_map:
            gradients = saliency_map(in_tensor, model, tm.output_name(), tm.channel_map[channel])
            plot_saliency_maps(in_tensor, gradients, paths, os.path.join(args.output_folder, f'{args.id}/saliency_maps/{tm.name}_{channel}'))


def _predict_and_evaluate(
    model, test_data, test_labels, tensor_maps_in, tensor_maps_out, tensor_maps_protected,
    batch_size, hidden_layer, plot_path, test_paths, embed_visualization, alpha, dpi, width, height,
):
    layer_names = [layer.name for layer in model.layers]
    performance_metrics = {}
    scatters = []
    rocs = []

    y_predictions = model.predict(test_data, batch_size=batch_size, verbose=0)
    protected_data = {tm: test_labels[tm.output_name()] for tm in tensor_maps_protected}
    for y, tm in zip(y_predictions, tensor_maps_out):
        if tm.output_name() not in layer_names:
            continue
        if isinstance(y_predictions, dict):
            y_predictions = y_predictions[tm.output_name()]
        if not isinstance(y_predictions, list):  # When models have a single output model.predict returns a ndarray otherwise it returns a list
            y = y_predictions
        y_truth = np.array(test_labels[tm.output_name()])
        performance_metrics.update(
            evaluate_predictions(
                tm, y, y_truth, protected_data, tm.name, plot_path, test_paths,
                rocs=rocs, scatters=scatters, dpi=dpi, width=width, height=height,
            ),
        )
        if tm.is_language():
            sample_from_language_model(tensor_maps_in[0], tm, model, test_data, max_samples=16)

    if len(rocs) > 1:
        subplot_rocs(rocs, plot_path)
    if len(scatters) > 1:
        subplot_scatters(scatters, plot_path)

    test_labels_1d = {tm: np.array(test_labels[tm.output_name()]) for tm in tensor_maps_out if tm.output_name() in test_labels}
    test_labels_1d.update(protected_data)
    if embed_visualization == "tsne":
        _tsne_wrapper(
            model, hidden_layer, alpha, plot_path, test_paths, test_labels_1d, test_data=test_data,
            tensor_maps_in=tensor_maps_in, batch_size=batch_size, dpi=dpi, width=width, height=height,
        )

    return performance_metrics


def _predict_scalars_and_evaluate_from_generator(
    model, generate_test, tensor_maps_in, tensor_maps_out, tensor_maps_protected,
    steps, hidden_layer, plot_path, alpha, dpi, width, height,
):
    layer_names = [layer.name for layer in model.layers]
    model_predictions = [tm.output_name() for tm in tensor_maps_out if tm.output_name() in layer_names]
    scalar_predictions = {tm.output_name(): [] for tm in tensor_maps_out if len(tm.shape) == 1 and tm.output_name() in layer_names}
    test_labels = {tm.output_name(): [] for tm in tensor_maps_out if len(tm.shape) == 1}
    protected_data = {tm: [] for tm in tensor_maps_protected}

    logging.info(f"Scalar predictions {model_predictions} names: {scalar_predictions.keys()} test labels: {test_labels.keys()}")
    embeddings = []
    test_paths = []
    for i in range(steps):
        batch = next(iter(generate_test))
        if len(batch) == 3:
            input_data, output_data, tensor_paths = batch[BATCH_INPUT_INDEX], batch[BATCH_OUTPUT_INDEX], batch[BATCH_PATHS_INDEX]
            test_paths.extend(tensor_paths)
        else:
            input_data, output_data = batch[BATCH_INPUT_INDEX], batch[BATCH_OUTPUT_INDEX]
            tensor_paths = None
        y_predictions = model.predict(input_data, verbose=0)
        if hidden_layer in layer_names:
            x_embed = embed_model_predict(model, tensor_maps_in, hidden_layer, input_data, 2)
            embeddings.extend(np.copy(np.reshape(x_embed, (x_embed.shape[0], np.prod(x_embed.shape[1:])))))

        for tm_output_name in test_labels:
            test_labels[tm_output_name].extend(np.copy(output_data[tm_output_name]))
        for tm in tensor_maps_protected:
            protected_data[tm].extend(np.copy(output_data[tm.output_name()]))

        for y, tm_output_name in zip(y_predictions, model_predictions):
            if not isinstance(y_predictions, list):  # When models have a single output model.predict returns a ndarray otherwise it returns a list
                y = y_predictions
            if tm_output_name in scalar_predictions:
                scalar_predictions[tm_output_name].extend(np.copy(y[tm_output_name]))

        if i % 100 == 0:
            logging.info(f'Processed {i} batches, {len(test_paths)} tensors.')

    performance_metrics = {}
    scatters = []
    rocs = []
    for tm in tensor_maps_protected:
        protected_data[tm] = np.array(protected_data[tm])

    for tm in tensor_maps_out:
        if tm.output_name() in scalar_predictions:
            y_predict = np.array(scalar_predictions[tm.output_name()])
            y_truth = np.array(test_labels[tm.output_name()])
            metrics = evaluate_predictions(
                tm, y_predict, y_truth, protected_data, tm.name, plot_path, test_paths,
                rocs=rocs, scatters=scatters, dpi=dpi, width=width, height=height,
            )
            performance_metrics.update(metrics)

    if len(rocs) > 1:
        subplot_rocs(rocs, plot_path)
    if len(scatters) > 1:
        subplot_scatters(scatters, plot_path)
    if len(embeddings) > 0:
        test_labels_1d = {tm: np.array(test_labels[tm.output_name()]) for tm in tensor_maps_out if tm.output_name() in test_labels}
        _tsne_wrapper(
            model, hidden_layer, alpha, plot_path, test_paths, test_labels_1d,
            embeddings=embeddings, dpi=dpi, width=width, height=height,
        )

    return performance_metrics


def _get_common_outputs(models_inputs_outputs, output_prefix):
    """Returns a set of outputs common to all the models so we can compare the models according to those outputs only"""
    all_outputs = []
    for (_, ios) in models_inputs_outputs.items():
        outputs = {k: v for (k, v) in ios.items() if k == output_prefix}
        for (_, output) in outputs.items():
            all_outputs.append(set(output))
    logging.info(f'Finding model output intersection of {all_outputs}')
    return reduce(set.intersection, all_outputs)


def _get_predictions(args, models_inputs_outputs, input_data, outputs, input_prefix, output_prefix):
    """Makes multi-modal predictions for a given number of models.

    Returns:
        dict: The nested dictionary of predicted values.

            {
                'tensor_map_1':
                    {
                        'model_1': [[value1, value2], [...]],
                        'model_2': [[value3, value4], [...]]
                    },
                'tensor_map_2':
                    {
                        'model_2': [[value5, value6], [...]],
                        'model_3': [[value7, value8], [...]]
                    }
            }
    """
    predictions = defaultdict(dict)
    for model_file in models_inputs_outputs.keys():
        args.tensor_maps_out = models_inputs_outputs[model_file][output_prefix]
        args.tensor_maps_in = models_inputs_outputs[model_file][input_prefix]
        args.model_file = model_file
        model = legacy_multimodal_multitask_model(**args.__dict__)
        model_name = os.path.basename(model_file).replace(MODEL_EXT, '_')

        # We can feed 'model.predict()' the entire input data because it knows what subset to use
        y_pred = model.predict(input_data, batch_size=args.batch_size)
        for i, tm in enumerate(args.tensor_maps_out):
            if tm in outputs:
                if len(args.tensor_maps_out) == 1:
                    predictions[tm][model_name] = y_pred
                else:
                    predictions[tm][model_name] = y_pred[i]

    return predictions


def _scalar_predictions_from_generator(args, models_inputs_outputs, generator, steps, outputs, input_prefix, output_prefix):
    """Makes multi-modal scalar predictions for a given number of models.

    Returns:
        dict: The nested dictionary of predicted values.

            {
                'tensor_map_1':
                    {
                        'model_1': [[value1, value2], [...]],
                        'model_2': [[value3, value4], [...]]
                    },
                'tensor_map_2':
                    {
                        'model_2': [[value5, value6], [...]],
                        'model_3': [[value7, value8], [...]]
                    }
            }
    """
    models = {}
    test_paths = []
    scalar_predictions = {}
    test_labels = {tm.output_name(): [] for tm in args.tensor_maps_out if len(tm.shape) == 1}

    for model_file in models_inputs_outputs:
        args.model_file = model_file
        args.tensor_maps_in = models_inputs_outputs[model_file][input_prefix]
        args.tensor_maps_out = models_inputs_outputs[model_file][output_prefix]
        model = legacy_multimodal_multitask_model(**args.__dict__)
        model_name = os.path.basename(model_file).replace(MODEL_EXT, '')
        models[model_name] = model
        scalar_predictions[model_name] = [tm for tm in models_inputs_outputs[model_file][output_prefix] if len(tm.shape) == 1]

    predictions = defaultdict(dict)
    for j in range(steps):
        batch = next(generator)
        input_data, output_data, tensor_paths = batch[BATCH_INPUT_INDEX], batch[BATCH_OUTPUT_INDEX], batch[BATCH_PATHS_INDEX]
        test_paths.extend(tensor_paths)
        for tl in test_labels:
            test_labels[tl].extend(np.copy(output_data[tl]))

        for model_name, model_file in zip(models, models_inputs_outputs):
            # We can feed 'model.predict()' the entire input data because it knows what subset to use
            y_predictions = models[model_name].predict(input_data)

            for y, tm in zip(y_predictions, models_inputs_outputs[model_file][output_prefix]):
                if not isinstance(y_predictions, list):  # When models have a single output model.predict returns a ndarray otherwise it returns a list
                    y = y_predictions
                if j == 0 and tm in scalar_predictions[model_name]:
                    predictions[tm][model_name] = []
                if tm in scalar_predictions[model_name]:
                    predictions[tm][model_name].extend(np.copy(y))

    for tm in predictions:
        logging.info(f"{tm.output_name()} labels: {len(test_labels[tm.output_name()])}")
        test_labels[tm.output_name()] = np.array(test_labels[tm.output_name()])
        for m in predictions[tm]:
            logging.info(f"{tm.output_name()} model: {m} prediction length:{len(predictions[tm][m])}")
            predictions[tm][m] = np.array(predictions[tm][m])

    return predictions, test_labels, test_paths


def _calculate_and_plot_prediction_stats(args, predictions, outputs, paths):
    rocs = []
    scatters = []
    for tm in args.tensor_maps_out:
        if tm not in predictions:
            continue
        plot_title = tm.name+'_'+args.id
        plot_folder = os.path.join(args.output_folder, args.id)
        if tm.is_categorical() and tm.axes() == 1:
            for m in predictions[tm]:
                logging.info(f"{tm.name} channel map {tm.channel_map}\nsum truth = {np.sum(outputs[tm.output_name()], axis=0)}\nsum preds = {np.sum(predictions[tm][m], axis=0)}")
            plot_rocs(
                predictions[tm], outputs[tm.output_name()], tm.channel_map, plot_title, plot_folder,
                dpi=args.dpi, width=args.plot_width, height=args.plot_height,
            )
            rocs.append((predictions[tm], outputs[tm.output_name()], tm.channel_map))
        elif tm.is_categorical() and tm.axes() == 4:
            for p in predictions[tm]:
                y = predictions[tm][p]
                melt_shape = (y.shape[0]*y.shape[1]*y.shape[2]*y.shape[3], y.shape[4])
                predictions[tm][p] = y.reshape(melt_shape)

            y_truth = outputs[tm.output_name()].reshape(melt_shape)
            plot_rocs(
                predictions[tm], y_truth, tm.channel_map, plot_title, plot_folder,
                dpi=args.dpi, width=args.plot_width, height=args.plot_height,
            )
            plot_precision_recalls(
                predictions[tm], y_truth, tm.channel_map, plot_title, plot_folder,
                dpi=args.dpi, width=args.plot_width, height=args.plot_height,
            )
            roc_aucs = get_roc_aucs(predictions[tm], y_truth, tm.channel_map)
            precision_recall_aucs = get_precision_recall_aucs(predictions[tm], y_truth, tm.channel_map)
            aucs = {"ROC": roc_aucs, "Precision-Recall": precision_recall_aucs}
            log_aucs(**aucs)
        elif tm.is_categorical() and tm.axes() == 3:
            # have to plot dice before the reshape
            plot_dice(
                predictions[tm], outputs[tm.output_name()], tm.channel_map, paths, plot_title, plot_folder,
                dpi=args.dpi, width=args.plot_width, height=args.plot_height,
            )
            for p in predictions[tm]:
                y = predictions[tm][p]
                melt_shape = (y.shape[0]*y.shape[1]*y.shape[2], y.shape[3])
                predictions[tm][p] = y.reshape(melt_shape)
            y_truth = outputs[tm.output_name()].reshape(melt_shape)
            plot_rocs(
                predictions[tm], y_truth, tm.channel_map, plot_title, plot_folder,
                dpi=args.dpi, width=args.plot_width, height=args.plot_height,
            )
            plot_precision_recalls(
                predictions[tm], y_truth, tm.channel_map, plot_title, plot_folder,
                dpi=args.dpi, width=args.plot_width, height=args.plot_height,
            )
            roc_aucs = get_roc_aucs(predictions[tm], y_truth, tm.channel_map)
            precision_recall_aucs = get_precision_recall_aucs(predictions[tm], y_truth, tm.channel_map)
            aucs = {"ROC": roc_aucs, "Precision-Recall": precision_recall_aucs}
            log_aucs(**aucs)
        elif tm.is_continuous() and tm.axes() == 1:
            scaled_predictions = {k: tm.rescale(predictions[tm][k]) for k in predictions[tm]}
            plot_scatters(
                scaled_predictions, tm.rescale(outputs[tm.output_name()]), plot_title, plot_folder,
                dpi=args.dpi, width=args.plot_width, height=args.plot_height,
            )
            scatters.append((scaled_predictions, tm.rescale(outputs[tm.output_name()]), plot_title, None))
            coefs = get_pearson_coefficients(scaled_predictions, tm.rescale(outputs[tm.output_name()]))
            log_pearson_coefficients(coefs, tm.name)
        elif tm.is_time_to_event():
            new_predictions = {}
            for m in predictions[tm]:
                c_index = concordance_index_censored(outputs[tm.output_name()][:, 0] == 1.0, outputs[tm.output_name()][:, 1], predictions[tm][m][:, 0])
                concordance_return_values = ['C-Index', 'Concordant Pairs', 'Discordant Pairs', 'Tied Predicted Risk', 'Tied Event Time']
                logging.info(f"Model: {m} {[f'{label}: {value}' for label, value in zip(concordance_return_values, c_index)]}")
                new_predictions[f'{m}_C_Index_{c_index[0]:0.3f}'] = predictions[tm][m]
            plot_rocs(
                new_predictions, outputs[tm.output_name()][:, 0, np.newaxis], {f'_vs_ROC': 0}, plot_title,
                plot_folder, dpi=args.dpi, width=args.plot_width, height=args.plot_height,
            )
            rocs.append((new_predictions, outputs[tm.output_name()][:, 0, np.newaxis], {f'_vs_ROC': 0}))
            plot_prediction_calibrations(
                new_predictions, outputs[tm.output_name()][:, 0, np.newaxis], {f'_vs_ROC': 0},
                plot_title, plot_folder, dpi=args.dpi,
                width=args.plot_width, height=args.plot_height,
            )
        elif tm.is_survival_curve():
            for m in predictions[tm]:
                plot_survival(
                    predictions[tm][m], outputs[tm.output_name()], f'{m}_{plot_title}', tm.days_window,
                    prefix=plot_folder, dpi=args.dpi, width=args.plot_width, height=args.plot_height,
                )
        else:
            scaled_predictions = {k: tm.rescale(predictions[tm][k]) for k in predictions[tm]}
            plot_scatters(
                scaled_predictions, tm.rescale(outputs[tm.output_name()]), plot_title, plot_folder,
                dpi=args.dpi, width=args.plot_width, height=args.plot_height,
            )
            coefs = get_pearson_coefficients(scaled_predictions, tm.rescale(outputs[tm.output_name()]))
            log_pearson_coefficients(coefs, tm.name)

    if len(rocs) > 1:
        subplot_comparison_rocs(rocs, plot_folder, dpi=args.dpi, width=args.plot_width, height=args.plot_height)
    if len(scatters) > 1:
        subplot_comparison_scatters(scatters, plot_folder, dpi=args.dpi, width=args.plot_width, height=args.plot_height)


def _tsne_wrapper(
    model, hidden_layer_name, alpha, plot_path, test_paths, test_labels, test_data=None,
    tensor_maps_in=None, batch_size=16, embeddings=None, dpi=300, width=7, height=7,
):
    """Plot 2D t-SNE of a model's hidden layer colored by many different co-variates.

    Callers must provide either model's embeddings or test_data on which embeddings will be inferred

    :param model: Keras model
    :param hidden_layer_name: String name of the hidden layer whose embeddings will be visualized
    :param alpha: Transparency of each data point
    :param plot_path: Image file name and path for the t_SNE plot
    :param test_paths: Paths for hd5 file containing each sample
    :param test_labels: Dictionary mapping TensorMaps to numpy arrays of labels (co-variates) to color code the t-SNE plots with
    :param test_data: Input data for the model necessary if embeddings is None
    :param tensor_maps_in: Input TensorMaps of the model necessary if embeddings is None
    :param batch_size: Batch size necessary if embeddings is None
    :param embeddings: (optional) Model's embeddings
    :return: None
    """
    if hidden_layer_name not in [layer.name for layer in model.layers]:
        logging.warning(f"Can't compute t-SNE, layer:{hidden_layer_name} not in provided model.")
        return

    if embeddings is None:
        embeddings = embed_model_predict(model, tensor_maps_in, hidden_layer_name, test_data, batch_size)

    gene_labels = []
    label_dict, categorical_labels, continuous_labels = test_labels_to_label_map(test_labels, len(test_paths))
    if len(categorical_labels) > 0 or len(continuous_labels) > 0 or len(gene_labels) > 0:
        plot_tsne(
            embeddings, categorical_labels, continuous_labels, gene_labels,
            label_dict, plot_path, alpha, dpi, width, height,
        )


if __name__ == '__main__':
    arguments = parse_args()
    run(arguments)  # back to the top



import argparse
import os
import csv
from getpass import getuser
from collections import defaultdict

JOIN_CHAR = '_'
SCRIPT_PATH = 'tensormap/mgb/tensor_maps_partners_ecg_labels.py'
TENSOR_FROM_FILE_MAKER = "make_partners_ecg_label"
PREFIX = "partners_ecg"
TENSOR_PATH_PREFIX = "partners_ecg_rest"


def _clean_label_string(string):
    """Replace spaces and slashes with JOIN_CHAR,
       and remove parentheses and commas"""
    string = string.replace(' ', JOIN_CHAR)
    string = string.replace('/', JOIN_CHAR)
    string = string.replace('(', '')
    string = string.replace(')', '')
    string = string.replace(',', '')
    return string


def _write_tmap_to_py(write_imports, py_file, label_maps, channel_maps, keys_in_hd5):
    """Given label_maps (which associates labels with source phrases)
       and channel_maps (which associates labels with unique sublabels),
       define the tensormaps to associate source phrases with precise labels,
       and write these maps in a python script
    """

    # Add import statements to .py
    if write_imports:
        py_file.write(f"from ml4h.TensorMap import TensorMap, Interpretation\n")
        py_file.write(f"from ml4h.tensormap.mgb.ecg import {TENSOR_FROM_FILE_MAKER}\n\n")

    for label in label_maps:
        cm = '{'

        for i, channel in enumerate(channel_maps[label]):
            cm += f"'{channel}': {i}, "

        # At this point, i = len(channel_maps[label])-1
        # If 'unspecified' is not a label, we need to add and index it
        if 'unspecified' not in channel_maps[label]:
            cm += f"'unspecified': {i+1}"

        cm += '}'

        for key in keys_in_hd5:
            py_file.write(f"{PREFIX}_{key}_{label} = TensorMap('{PREFIX}_{key}_{label}', interpretation=Interpretation.CATEGORICAL, time_series_limit=0, path_prefix='{TENSOR_PATH_PREFIX}', channel_map={cm}, tensor_from_file={TENSOR_FROM_FILE_MAKER}(keys='{key}', dict_of_list = {label_maps[label]})) \n\n")
            py_file.write(f"{PREFIX}_{key}_{label}_newest = TensorMap('{PREFIX}_{key}_{label}_newest', interpretation=Interpretation.CATEGORICAL, path_prefix='{TENSOR_PATH_PREFIX}', channel_map={cm}, tensor_from_file={TENSOR_FROM_FILE_MAKER}(keys='{key}', dict_of_list = {label_maps[label]})) \n\n")


def _write_partners_ecg_tmap_script(py_file, partners_ecg_label_dir, keys_in_hd5):
    # Set flag for writing import statements to true
    write_imports = True

    # Iterate through all files in the mgb CSV labels folder
    for file in os.listdir(partners_ecg_label_dir):

        # Ignore files that do not end in .csv
        if not file.endswith('.csv'):
            continue

        # Isolate the task name
        task = file.replace('.csv', '').replace('c_', '')

        # Create list of lists;
        # outer list is rows in CSV,
        # inner list is columns; idx 0 is source phrase
        #                        idx 1 is label (level 1 of hierarchy)
        #                        idx 2 is label (level 2 of hierarchy) etc.
        fpath_this_task = os.path.join(partners_ecg_label_dir, file)
        lol = list(csv.reader(open(fpath_this_task), delimiter=','))

        # Associate labels with source phrases in dict of dicts:
        #   keys   - task name and all oot-level labels in hierarchy
        #   values - dicts:
        #       keys   - labels (next level down in hierarchy)
        #       values - list of source phrases that map to a given label
        # Note: because the first key is the task name, keys of dicts in
        # label_map[task] are the remaining keys in label_map itself
        label_maps = defaultdict(dict)

        # Associate labels with unique set of sublabels in dict of sets
        # keys   - every label in hierarchy with children
        # values - set of all child labels within a given label
        channel_maps = defaultdict(set)

        # Iterate through every source phrase in list of lists (label map)
        for row in lol:

            # Initialize prefix as empty list
            prefix = []

            # For the row, iterate through label strings 2:end
            for label_str in row[1:]:

                # If the label string is blank, skip
                if label_str == '':
                    continue

                # Clean label string
                label_str = _clean_label_string(label_str)

                # If ???
                if len(prefix) == 0:
                    channel_maps[task].add(label_str)

                    # If we already have a list of source phrases for this
                    # task and label string, append the source phrase to it
                    if label_str in label_maps[task]:
                        label_maps[task][label_str].append(row[0])

                    # If no list exists yet for this task and label string,
                    # create a new list and initialize with the source phrase
                    else:
                        label_maps[task][label_str] = [row[0]]
                else:
                    # Join elements in prefix list by JOIN_CHAR
                    prefix_merged = JOIN_CHAR.join(prefix)
                    channel_maps[prefix_merged].add(label_str)

                    if label_str in label_maps[prefix_merged]:
                        label_maps[prefix_merged][label_str].append(row[0])
                    else:
                        label_maps[prefix_merged][label_str] = [row[0]]

                    label_maps[prefix_merged][label_str].append(row[0])

                prefix.append(label_str)

        _write_tmap_to_py(write_imports, py_file, label_maps, channel_maps, keys_in_hd5)
        write_imports = False

        print(f"Created Tensor Maps from {file} and saved in {SCRIPT_PATH}")


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--label_maps_dir', default=f'/home/{getuser()}/partners_ecg_labeling', help='Path to directory with c_task.csv label maps.')
    parser.add_argument('--hd5_keys', default=['read_md_clean', 'read_pc_clean'], nargs='+')
    args = parser.parse_args()

    # Open the python script, parse the .csv label maps,
    # and create code that creates TensorMaps
    with open(SCRIPT_PATH, 'w') as py_file:
        _write_partners_ecg_tmap_script(py_file, args.label_maps_dir, args.hd5_keys)



import logging
from typing import Dict, List, Tuple

import numpy as np

import keras
import tensorflow as tf
import tensorflow.keras.backend as K
import tensorflow_probability as tfp
from tensorflow.keras.losses import categorical_crossentropy
from tensorflow.keras.layers import concatenate, Flatten, Average, Layer, Lambda, Dense

from ml4h.models.Block import Block
from ml4h.TensorMap import TensorMap
from ml4h.models.basic_blocks import DenseBlock
from ml4h.models.layer_wrappers import global_average_pool
from tensorflow.keras.losses import categorical_crossentropy
from keras.saving import register_keras_serializable


Tensor = tf.Tensor
tfd = tfp.distributions


class FlatConcatBlock(Block):
    """
    Flattens then concatenates all inputs
    """
    def __init__(self,  **kwargs):
        pass

    def __call__(self, x: Tensor, intermediates: Dict[TensorMap, List[Tensor]] = None) -> Tensor:
        y = [Flatten()(x[-1]) for tm, x in intermediates.items()]
        y = concatenate(y) if len(y) > 1 else y[0]
        return y


class FlatConcatDenseBlock(Block):
    """
    Flattens then concatenates all inputs, applies a dense layer
    """
    def __init__(
            self,
            activation: str = 'swish',
            dense_layers: List[int] = [32],
            dense_normalize: str = None,
            dense_regularize: str = None,
            dense_regularize_rate: float = 0.0,
            **kwargs,
    ):
        self.fully_connected = DenseBlock(
            widths=dense_layers,
            activation=activation,
            normalization=dense_normalize,
            regularization=dense_regularize,
            regularization_rate=dense_regularize_rate,
            name='embed',
        ) if dense_layers else None

    def __call__(self, x: Tensor, intermediates: Dict[TensorMap, List[Tensor]] = None) -> Tensor:
        y = [Flatten()(x[-1]) for tm, x in intermediates.items()]
        y = concatenate(y) if len(y) > 1 else y[0]
        y = self.fully_connected(y, intermediates) if self.fully_connected else y
        return y


class KroneckerBlock(Block):
    """
    Outer-product of all paired inputs, applies a dense layer
    """
    def __init__(
            self,
            activation: str = 'swish',
            dense_layers: List[int] = [32],
            dense_normalize: str = None,
            dense_regularize: str = None,
            dense_regularize_rate: float = 0.0,
            pairs: List[Tuple[TensorMap, TensorMap]] = None,
            **kwargs,
    ):
        self.pairs = pairs
        self.encoding_size = dense_layers[-1]
        self.fully_connected = DenseBlock(
            widths=dense_layers,
            activation=activation,
            normalization=dense_normalize,
            regularization=dense_regularize,
            regularization_rate=dense_regularize_rate,
            name='embed',
        ) if dense_layers else None

    def __call__(self, x: Tensor, intermediates: Dict[TensorMap, List[Tensor]] = None) -> Tensor:
        y = [Flatten()(x[-1]) for tm, x in intermediates.items()]
        eshape = tf.shape(intermediates.items()[0][-1])
        if len(y) == 2:
            logging.info(f'********\n\n\n*********** Trying KRONECKER {self.encoding_size}\n*******************\n\n')
            kron_layer = Lambda(lambda tensors: tf.einsum('...i,...j->...ij', y[0], y[1]))
            kron = kron_layer(y)
            y = tf.reshape(kron, [eshape[0], self.encoding_size * self.encoding_size])

        y = self.fully_connected(y, intermediates) if self.fully_connected else y
        return y


class DropoutBlock(Block):
    """
    Dropout from embeddings of different modalities
    """

    def __init__(
            self,
            activation: str = 'swish',
            dense_layers: List[int] = [256],
            dense_normalize: str = None,
            dense_regularize: str = None,
            dense_regularize_rate: float = 0.0,
            pairs: List[Tuple[TensorMap, TensorMap]] = None,
            **kwargs,
    ):
        self.pairs = pairs
        self.encoding_size = dense_layers[-1]
        self.fully_connected = DenseBlock(
            widths=dense_layers,
            activation=activation,
            normalization=dense_normalize,
            regularization=dense_regularize,
            regularization_rate=dense_regularize_rate,
            name='embed',
        ) if dense_layers else None

    def __call__(self, x: Tensor, intermediates: Dict[TensorMap, List[Tensor]] = None) -> Tensor:
        for left, right in self.pairs:
            random_index = tf.random.uniform(
                shape=[intermediates[left][-1].shape[-1]],
                maxval=len(intermediates), dtype=tf.int32,
            )
            ranger = tf.range(intermediates[left][-1].shape[-1])
            indices = tf.stack([random_index, ranger], axis=-1)
            tf_y = tf.convert_to_tensor([intermediates[left][-1], intermediates[right][-1]])
            tf_y = tf.transpose(tf_y, perm=[0, 2, 1])
            tf_g = tf.gather_nd(tf_y, indices)
            y = tf.transpose(tf_g)
            y = self.fully_connected(y, intermediates) if self.fully_connected else y
            return y


        # elif self.pair_merge == 'dropout':
        #     random_index = tf.random.uniform(shape=[intermediates[left][-1].shape[-1]], maxval=len(y), dtype=tf.int32)
        #     ranger = tf.range(intermediates[left][-1].shape[-1])
        #     indices = tf.stack([random_index, ranger], axis=-1)
        #     tf_y = tf.convert_to_tensor(y)
        #     tf_y = tf.transpose(tf_y, perm=[0, 2, 1])
        #     tf_g = tf.gather_nd(tf_y, indices)
        #     out = tf.transpose(tf_g)
        #     return out


class GlobalAveragePoolBlock(Block):
    """
    GAPs then concatenates all inputs, applies a dense layer
    """
    def __init__(
            self,
            activation: str = 'swish',
            dense_layers: List[int] = [32],
            dense_normalize: str = None,
            dense_regularize: str = None,
            dense_regularize_rate: float = 0.0,
            **kwargs,
    ):
        self.fully_connected = DenseBlock(
            widths=dense_layers,
            activation=activation,
            normalization=dense_normalize,
            regularization=dense_regularize,
            regularization_rate=dense_regularize_rate,
            name='embed',
        ) if dense_layers else None

    def __call__(self, x: Tensor, intermediates: Dict[TensorMap, List[Tensor]] = None) -> Tensor:
        y = [Flatten()(x[-1]) for tm, x in intermediates.items() if tm.axes() == 1]  # Flat tensors
        y += [global_average_pool(x[-2 if self.fully_connected else -1]) for tm, x in intermediates.items() if tm.axes() > 1]  # Structured tensors
        y = concatenate(y) if len(y) > 1 else y[0]
        y = self.fully_connected(y, intermediates) if self.fully_connected else y
        return y


class AverageBlock(Block):
    """
    Average the last tensors in intermediates dictionary
    """
    def __init__(self, **kwargs):
        pass

    def __call__(self, x: Tensor, intermediates: Dict[TensorMap, List[Tensor]] = None) -> Tensor:
        return Average()([t[-1] for tm, t in intermediates.items()])


class ReduceMean(Block):
    """
    Average the last tensors in intermediates dictionary
    """
    def __init__(self, **kwargs):
        pass

    def __call__(self, x: Tensor, intermediates: Dict[TensorMap, List[Tensor]] = None) -> Tensor:
        y = [x[-1] for tm, x in intermediates.items()]
        y = tf.math.reduce_mean(y, axis=0)
        return y


class EncodeIdentityBlock(Block):
    """
    Adds the input tensor to the intermediates dictionary, useful for TensorMaps with pretrained embeddings
    """
    def __init__(self, **kwargs):
        super()

    def __call__(self, x: Tensor, intermediates: Dict[TensorMap, List[Tensor]] = None) -> Tensor:
        return x


class PairLossBlock(Block):
    """
    Flattens or GAPs then concatenates all inputs, applies a dense layer, then restructures to provided shapes
    """
    def __init__(
            self,
            pairs: List[Tuple[TensorMap, TensorMap]],
            pair_loss: str = 'cosine',
            pair_loss_weight: float = 1.0,
            pair_merge: str = 'dropout',
            batch_size: int = 4,
            dense_layers: List[int] = [32],
            **kwargs,
    ):
        self.pairs = pairs
        self.pair_merge = pair_merge
        self.batch_size = batch_size
        self.encoding_size = dense_layers[-1]
        if pair_loss == 'cosine':
            self.loss_layer = CosineLossLayer(pair_loss_weight)
        elif pair_loss == 'euclid':
            self.loss_layer = L2LossLayer(pair_loss_weight)
        elif pair_loss == 'contrastive':
            self.loss_layer = ContrastiveLossLayer(pair_loss_weight, batch_size)
        else:
            raise ValueError(f'Unknown pair loss type: {pair_loss}')

    def get_config(self):
        return self.loss_layer.get_config()

    #@tf.function()
    def __call__(self, x: Tensor, intermediates: Dict[TensorMap, List[Tensor]] = None) -> Tensor:
        y = []
        for left, right in self.pairs:
            y.extend(self.loss_layer([intermediates[left][-1], intermediates[right][-1]]))
        if self.pair_merge == 'average':
            return Average()(y)
        elif self.pair_merge == 'concat':
            return concatenate(y)
        elif self.pair_merge == 'dropout':
            return DropoutMergeLayer()(y)
        elif self.pair_merge == 'kronecker':
            krons = []
            for left, right in self.pairs:
                eshape = tf.shape(intermediates[left][-1])
                kron_layer = Lambda(lambda tensors: tf.einsum('...i,...j->...ij', tensors[0], tensors[1]))
                y = self.loss_layer([intermediates[left][-1], intermediates[right][-1]])
                kron = kron_layer(y)
                krons.append(tf.reshape(kron, [eshape[0], self.encoding_size*self.encoding_size]))
            if len(self.pairs) > 1:
                kron = concatenate(krons)
            else:
                kron = krons[0]
            kron = Dense(self.encoding_size)(kron)
            return kron
        else:
            raise ValueError(f'Unknown pair merge method: {self.pair_merge}')
    def _dropout_merge(self, y):
        tf_y = tf.stack(y, axis=0)  # shape: [num_pairs, batch, dim]
        tf_y = tf.transpose(tf_y, perm=[1, 2, 0])  # shape: [batch, dim, num_pairs]

        batch_size = tf.shape(tf_y)[0]
        dim = tf.shape(tf_y)[1]
        num_pairs = tf.shape(tf_y)[2]

        random_indices = tf.random.uniform(shape=[batch_size], maxval=num_pairs, dtype=tf.int32)
        dim_range = tf.range(dim)
        dim_range = tf.reshape(dim_range, [1, -1])
        dim_range = tf.tile(dim_range, [batch_size, 1])

        batch_indices = tf.range(batch_size)
        batch_indices = tf.reshape(batch_indices, [batch_size, 1])
        batch_indices = tf.tile(batch_indices, [1, dim])
        gather_indices = tf.stack([batch_indices, dim_range], axis=-1)

        gather_indices = tf.reshape(gather_indices, [-1, 2])
        rand_idx_exp = tf.reshape(tf.repeat(random_indices, dim), [-1, 1])
        gather_indices = tf.concat([gather_indices, rand_idx_exp], axis=-1)

        gathered = tf.gather_nd(tf_y, gather_indices)
        return tf.reshape(gathered, [batch_size, dim])

@register_keras_serializable()
class DropoutMergeLayer(tf.keras.layers.Layer):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
    def call(self, y):
        tf_y = tf.stack(y, axis=0)  # shape: [num_pairs, batch, dim]
        tf_y = tf.transpose(tf_y, perm=[1, 2, 0])  # shape: [batch, dim, num_pairs]

        batch_size = tf.shape(tf_y)[0]
        dim = tf.shape(tf_y)[1]
        num_pairs = tf.shape(tf_y)[2]

        random_indices = tf.random.uniform(shape=[batch_size], maxval=num_pairs, dtype=tf.int32)
        dim_range = tf.range(dim)
        dim_range = tf.reshape(dim_range, [1, -1])
        dim_range = tf.tile(dim_range, [batch_size, 1])

        batch_indices = tf.range(batch_size)
        batch_indices = tf.reshape(batch_indices, [batch_size, 1])
        batch_indices = tf.tile(batch_indices, [1, dim])
        gather_indices = tf.stack([batch_indices, dim_range], axis=-1)

        gather_indices = tf.reshape(gather_indices, [-1, 2])
        rand_idx_exp = tf.reshape(tf.repeat(random_indices, dim), [-1, 1])
        gather_indices = tf.concat([gather_indices, rand_idx_exp], axis=-1)

        gathered = tf.gather_nd(tf_y, gather_indices)
        return tf.reshape(gathered, [batch_size, dim])
    def get_config(self):
        config = super().get_config()
        return config


def l2_norm(x, axis=None):
    """
    takes an input tensor and returns the l2 norm along specified axis
    """

    square_sum = K.sum(K.square(x), axis=axis, keepdims=True)
    norm = K.sqrt(K.maximum(square_sum, K.epsilon()))

    return norm


def pairwise_cosine_difference(t1, t2):
    t1_norm = t1 / l2_norm(t1, axis=-1)
    t2_norm = t2 / l2_norm(t2, axis=-1)
    dot = K.clip(K.batch_dot(t1_norm, t2_norm), -1, 1)
    return K.mean(tf.acos(dot))


def contrastive_difference(left: Tensor, right: Tensor, batch_size: int, temperature: Tensor):
    left_normed = left / l2_norm(left, axis=-1)
    right_normed = right / l2_norm(right, axis=-1)
    logits_left = tf.linalg.matmul(left_normed, right_normed, transpose_b=True) * tf.math.exp(temperature)
    logits_right = tf.linalg.matmul(right_normed, left_normed, transpose_b=True) * tf.math.exp(temperature)
    prob_left = tf.keras.activations.softmax(logits_left, axis=-1)
    prob_right = tf.keras.activations.softmax(logits_right, axis=-1)

    # identity matrix (np.eye) matches left row modality with right column modality
    labels = tf.convert_to_tensor(np.eye(batch_size), dtype=tf.float32)
    loss_left = tf.keras.losses.CategoricalCrossentropy(from_logits=False, reduction=tf.keras.losses.Reduction.SUM)(prob_left, labels)
    loss_right = tf.keras.losses.CategoricalCrossentropy(from_logits=False, reduction=tf.keras.losses.Reduction.SUM)(prob_right, labels)
    loss = (loss_left + loss_right)/2
    return loss / batch_size

@register_keras_serializable()
class CosineLossLayer(Layer):
    """Layer that creates an Cosine loss."""

    def __init__(self, weight, **kwargs):
        super(CosineLossLayer, self).__init__(**kwargs)
        self.weight = weight

    def get_config(self):
        config = super().get_config().copy()
        config.update({'weight': self.weight})
        return config

    def call(self, inputs):
        # We use `add_loss` to create a regularization loss
        # that depends on the inputs.
        self.add_loss(self.weight * pairwise_cosine_difference(inputs[0], inputs[1]))
        return inputs

@register_keras_serializable()
class L2LossLayer(Layer):
    """Layer that creates an L2 loss."""

    def __init__(self, weight, **kwargs):
        super(L2LossLayer, self).__init__(**kwargs)
        self.weight = weight

    def get_config(self):
        config = super().get_config().copy()
        config.update({'weight': self.weight})
        return config

    def call(self, inputs):
        #self.add_loss(self.weight * tf.reduce_sum(tf.square(inputs[0] - inputs[1])))
        self.add_loss(self.weight * K.mean(l2_norm(inputs[0] - inputs[1])))
        return inputs


@register_keras_serializable()
class ContrastiveLossLayer(Layer):
    """Layer that creates an Cosine loss."""

    def __init__(self, weight, batch_size, **kwargs):
        super(ContrastiveLossLayer, self).__init__(**kwargs)
        self.weight = weight
        self.batch_size = batch_size
        self.temperature = self.add_weight(
            name='contrastive_temperature',
            shape=(1,), initializer="zeros", trainable=True,
        )

    def get_config(self):
        config = super().get_config().copy()
        config.update({'weight': self.weight, 'batch_size': self.batch_size})
        return config

    def call(self, inputs, training = None):
        # We use `add_loss` to create a regularization loss
        # that depends on the inputs.
        # if training is False:
        #     return inputs
        contrastive_loss = self.weight * contrastive_difference(inputs[0], inputs[1], self.batch_size, self.temperature)
        self.add_loss(contrastive_loss)
        #self.add_metric(contrastive_loss)
        return inputs


class LinearTransform(tf.keras.layers.Layer):
    """Layer that implements y=m*x+b, where m and b are
    learnable parameters.
    """
    def __init__(
        self,
        gamma_initializer="ones",
        beta_initializer="zeros",
        dtype=None,
        **kwargs
    ):
        super().__init__(dtype=dtype, **kwargs)
        self.gamma_initializer = gamma_initializer
        self.beta_initializer = beta_initializer

    def build(self, input_shape):
        num_channels = int(input_shape[-1])
        self.gamma = self.add_weight(
            "gamma",
            shape=[num_channels],
            initializer=self.gamma_initializer,
            dtype=self.dtype,
        )
        self.beta = self.add_weight(
            "beta",
            shape=[num_channels],
            initializer=self.beta_initializer,
            dtype=self.dtype,
        )

    def call(self, inputs):
        return self.gamma * inputs[0] + self.beta


class VariationalDiagNormal(Layer):
    def __init__(
            self,
            latent_size: int,
            kl_divergence_weight: float = 1.,
            **kwargs
    ):
        self.latent_size = latent_size
        self.kl_divergence_weight = kl_divergence_weight
        super(VariationalDiagNormal, self).__init__(**kwargs)
        self.prior = tfd.MultivariateNormalDiag(loc=tf.zeros([latent_size]), scale_identity_multiplier=1.0)

    def call(self, mu: Tensor, log_sigma: Tensor, **kwargs):
        """mu and sigma must be shape (None, latent_size)"""
        approx_posterior = tfd.MultivariateNormalDiag(loc=mu, scale_diag=tf.math.exp(log_sigma))
        kl = tf.reduce_mean(tfd.kl_divergence(approx_posterior, self.prior))
        self.add_loss(kl * self.kl_divergence_weight)
        self.add_metric(kl, name='KL_divergence')
        return approx_posterior.sample()

    def get_config(self):
        return {'latent_size': self.latent_size, 'kl_divergence_weight': self.kl_divergence_weight}
